# Copyright 2019-2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import logging
import os

import attr
import yaml

from klio_core.config import _io as io
from klio_core.config import _preprocessing as preprocessing
from klio_core.config import _utils as utils

logger = logging.getLogger("klio")


# path used by both klio-cli and klioexec to write and read the config when
# running a job
RUN_EFFECTIVE_CONFIG_FILE = "klio-job-run-effective.yaml"
RUN_EFFECTIVE_CONFIG_PATH = os.path.join(
    "/usr/src/app", RUN_EFFECTIVE_CONFIG_FILE
)

WORKER_DISK_TYPE_URL = (
    "compute.googleapis.com/projects/{project}/regions/{region}/"
    "diskTypes/{disk_type}"
)


@utils.config_object(key_prefix=None, allow_unknown_keys=False)
class BaseKlioConfig(object):
    """Base Klio config representation of ``klio-job.yaml``.

    Args:
        config_dict (dict): dictionary representation of configuration
            as parsed from ``klio-job.yaml``.
    """

    job_name = attr.attrib(type=str)

    job_config = attr.attrib(repr=False)
    pipeline_options = attr.attrib(repr=False, default={})

    version = attr.attrib(type=int, default=2, repr=False)

    def __config_post_init__(self, config_dict):
        self._raw = config_dict

        self.job_config = KlioJobConfig(
            self.job_config, job_name=self.job_name, version=self.version,
        )
        self.pipeline_options = KlioPipelineConfig(
            self.pipeline_options,
            job_name=self.job_name,
            version=self.version,
        )

    def __repr__(self):
        return "KlioConfig(job_name=%s)" % self.job_name

    def as_dict(self):
        job_config_dict = self.job_config.as_dict()
        pipeline_opts_dict = self.pipeline_options.as_dict()
        base_config_dict = attr.asdict(self)
        base_config_dict["job_config"] = job_config_dict
        base_config_dict["pipeline_options"] = pipeline_opts_dict
        return base_config_dict


class KlioConfig(BaseKlioConfig):
    """Klio config object representation of ``klio-job.yaml``.

    Args:
        config_dict (dict): dictionary representation of configuration
            as parsed from ``klio-job.yaml``.

    Attributes:
        job_name (str): Name of Klio job.
        version (int): Version of Klio job.
        pipeline_options (KlioPipelineConfig): Apache Beam pipeline-related
            configuration.
        job_config (KlioJobConfig): Job-related configuration.
    """

    def __init__(
        self,
        raw_config_data,
        raw_overrides=None,
        raw_templates=None,
        config_skip_preprocessing=False,
        **kwargs
    ):
        if config_skip_preprocessing:
            processed_config = raw_config_data
        else:
            processed_config = preprocessing.KlioConfigPreprocessor.process(
                raw_config_data=raw_config_data,
                raw_template_list=raw_templates or [],
                raw_override_list=raw_overrides or [],
            )
        return super().__init__(processed_config, **kwargs)

    # re-defining as_dict just to have its docstring.
    def as_dict(self):
        """Return a dictionary representation of the :class:`KlioConfig`
        object.
        """
        return super().as_dict()

    def write_to_file(self, path_or_stream):
        """write the config object as a yaml file

        Args:
            path_or_stream: either a string of a file path or a writable stream
        """

        def _write(stream):
            yaml.dump(
                self.as_dict(),
                stream=stream,
                Dumper=utils.IndentListDumper,
                default_flow_style=False,
                sort_keys=False,
            )

        if isinstance(path_or_stream, str):
            dirname = os.path.dirname(path_or_stream)
            if not os.path.exists(dirname):
                os.makedirs(dirname)
            with open(path_or_stream, "w") as stream:
                _write(stream)
        else:
            _write(path_or_stream)


@attr.attrs
class KlioIOConfigContainer(object):
    inputs = attr.attrib()
    outputs = attr.attrib()


@utils.config_object(key_prefix="job_config", allow_unknown_keys=True)
class KlioJobConfig(object):
    """Job-specific config representing the ``job_config`` key of
    ``klio-job.yaml``.

    ``job_config`` is both for any user-specific job configuration needed,
    as well as klio-related configuration (i.e. timeouts, metrics).

    See :ref:`documentation <job-config>` for information on available
    configuration.

    Attributes:
        job_name (str): Name of Klio job.
        version (int): Version of Klio job.
        allow_non_klio_messages (bool): Allow this job to process free-form,
            non-KlioMessage messages.
        blocking (bool): Wait for Dataflow job to finish before exiting.
        metrics (dict): Dictionary representing desired metrics configuration.
        events (``KlioIOConfigContainer``): Job event I/O configuration.
        data (``KlioIOConfigContainer``): Job data I/O configuration.

    Args:
        config_dict (dict): dictionary representation of ``job_config``
            as parsed from ``klio-job.yaml``.
    """

    ATTRIBS_TO_SKIP = ["version", "job_name"]

    # required attributes
    job_name = utils.field(type=str, repr=True)
    version = utils.field(type=int)

    # optional attributes
    allow_non_klio_messages = utils.field(type=bool, default=False)
    metrics = utils.field(default={})
    blocking = utils.field(type=bool, default=False)

    def __config_post_init__(self, config_dict):
        self._raw = config_dict
        self._scanned_io_subclasses = None
        self.USER_ATTRIBS = []

        self._parse_io(config_dict)

        declared_config = self._as_dict()
        for key, value in self._raw.items():
            if key not in declared_config:
                # set user attributes to job_config - but only the first
                # level, for example, from this
                # job_config:
                #    foo:
                #      key1: value1
                #      list1:
                #        - one
                #        - two
                # gets parsed to:
                # job_config.foo -> {"key1": "value1", "alist": ["one", "two"]}
                setattr(self, key, value)
                # key track of user-set attributes so that when as_dict()
                # is called, we re-add it to what _as_dict() returns
                self.USER_ATTRIBS.append({key: value})

    def _parse_io(self, config_dict):
        event_inputs = self._create_config_objects(
            config_dict.get("events", {}).get("inputs", {}),
            io.KlioIOType.EVENT,
            io.KlioIODirection.INPUT,
        )
        event_outputs = self._create_config_objects(
            config_dict.get("events", {}).get("outputs", {}),
            io.KlioIOType.EVENT,
            io.KlioIODirection.OUTPUT,
        )
        self.events = KlioIOConfigContainer(
            inputs=event_inputs, outputs=event_outputs
        )

        data_inputs = self._create_config_objects(
            config_dict.get("data", {}).get("inputs", {}),
            io.KlioIOType.DATA,
            io.KlioIODirection.INPUT,
        )
        data_outputs = self._create_config_objects(
            config_dict.get("data", {}).get("outputs", {}),
            io.KlioIOType.DATA,
            io.KlioIODirection.OUTPUT,
        )
        self.data = KlioIOConfigContainer(
            inputs=data_inputs, outputs=data_outputs
        )

    def _get_all_config_subclasses(self):
        if self._scanned_io_subclasses is not None:
            return self._scanned_io_subclasses

        # notice this will end up including intermediate classes but that
        # shouldn't matter since they shouldn't "support" any valid
        # combinations of type/direction
        all_subclasses = []

        def traverse(cls):
            for subclass in cls.__subclasses__():
                all_subclasses.append(subclass)
                traverse(subclass)

        traverse(io.KlioIOConfig)
        self._scanned_io_subclasses = all_subclasses
        return all_subclasses

    def _create_config_objects(self, configs, io_type, io_direction):
        options = dict(
            [
                (x.name, x)
                for x in self._get_all_config_subclasses()
                if x.supports_type(io_type)
                and x.supports_direction(io_direction)
            ]
        )
        objs = []
        for name, config in configs.items():
            type_name = config["type"].lower()
            if type_name not in options:
                raise Exception(
                    "{} is not a valid {} {}".format(
                        config["type"], io_type.name, io_direction.name
                    )
                )
            subclass = options[type_name]
            objs.append(subclass.from_dict(config, io_type, io_direction))
        return objs

    def _as_dict(self):
        config_dict = attr.asdict(
            self, filter=lambda x, _: x.name not in self.ATTRIBS_TO_SKIP
        )
        config_dict["events"] = {}
        config_dict["events"]["inputs"] = [
            ei.as_dict() for ei in self.events.inputs
        ]
        config_dict["events"]["outputs"] = [
            eo.as_dict() for eo in self.events.outputs
        ]

        config_dict["data"] = {}
        config_dict["data"]["inputs"] = [
            di.as_dict() for di in self.data.inputs
        ]
        config_dict["data"]["outputs"] = [
            do.as_dict() for do in self.data.outputs
        ]
        return config_dict

    def as_dict(self):
        """Return a dictionary representation of the :class:`KlioJobConfig`
        object.

        .. tip::

            Use this method to access any custom config key/value pairs
            defined under ``klio-job.yaml::job_config``.
        """

        config_dict = self._as_dict()
        for attrib in self.USER_ATTRIBS:
            for key, value in attrib.items():
                config_dict[key] = value

        return config_dict

    def __repr__(self):
        return "KlioJobConfig(job_name=%s)" % self.job_name


@utils.config_object(key_prefix="pipeline_options", allow_unknown_keys=True)
class KlioPipelineConfig(object):
    """Pipeline-specific config representing the ``pipeline_options`` key
    of ``klio-job.yaml``.

    .. note::

        ``pipeline_options`` map 1:1 to options supported in Apache Beam and
        its runners (i.e. Dataflow). See `all supported pipeline options
        <https://github.com/apache/beam/blob/master/sdks/python/apache_beam/
        options/pipeline_options.py>`_ for available options.

    Any instance attribute not defined in this class but is available in Apache
    Beam or its runners will still be passed through when running the
    pipeline.

    See :ref:`documentation <kliopipelineconfig>` for information on available
    configuration.

    Args:
        config_dict (dict): dictionary representation of ``pipeline_options``
            as parsed from ``klio-job.yaml``.

    Attributes:
        job_name (str): Name of Klio job.
        version (int): Version of Klio job.
        (remaining attributes): See `all supported pipeline options
            <https://github.com/apache/beam/blob/master/sdks/python/
            apache_beam/options/pipeline_options.py>`_ for all available
            remaining attributes.
    """

    ATTRIBS_TO_SKIP = ["version"]

    job_name = utils.field(repr=True, type=str)
    version = utils.field(type=int)

    runner = utils.field(type=str, default="DataflowRunner")
    streaming = utils.field(type=bool, default=True)

    # debug options
    # TODO: add validation - if job is instreaming mode and no setup.py,
    # then config must use the "beam_fn_api" experiment (we could also just
    # automatically put it in there)
    experiments = utils.field(default=[])

    # setup options
    sdk_location = utils.field(type=str, default=None)
    setup_file = utils.field(type=str, default=None)
    requirements_file = utils.field(type=str, default=None)

    # GCP options
    project = utils.field(type=str, default=None)
    staging_location = utils.field(type=str, default=None)
    temp_location = utils.field(type=str, default=None)
    region = utils.field(type=str, default="europe-west1")
    subnetwork = utils.field(type=str, default=None)
    update = utils.field(type=bool, default=False)
    dataflow_endpoint = utils.field(type=str, default=None)
    service_account_email = utils.field(type=str, default=None)
    no_auth = utils.field(type=bool, default=False)
    template_location = utils.field(type=str, default=None)
    labels = utils.field(default=[])
    label = utils.field(type=str, default=None)
    transform_name_mapping = utils.field(type=str, default=None)
    enable_streaming_engine = utils.field(type=bool, default=False)
    dataflow_kms_key = utils.field(type=str, default=None)
    flexrs_goal = utils.field(
        type=str,
        default=None,
        validator=attr.validators.optional(
            attr.validators.in_(["COST_OPTIMIZED", "SPEED_OPTIMIZED"])
        ),
    )

    # worker options
    autoscaling_algorithm = utils.field(
        type=str,
        default="NONE",
        validator=attr.validators.optional(
            attr.validators.in_(["THROUGHPUT_BASED", "NONE"])
        ),
    )
    num_workers = utils.field(type=int, default=2)
    disk_size_gb = utils.field(type=int, default=32)
    worker_machine_type = utils.field(type=str, default="n1-standard-2")
    worker_harness_container_image = utils.field(type=str, default=None)
    worker_disk_type = utils.field(type=str, default=None,)
    use_public_ips = utils.field(type=bool, default=None)
    min_cpu_platform = utils.field(type=str, default=None)
    dataflow_worker_jar = utils.field(type=str, default=None)

    # profiling options
    profile_location = utils.field(type=str, default=None)
    profile_cpu = utils.field(type=bool, default=None)
    profile_memory = utils.field(type=bool, default=None)
    profile_sample_rate = utils.field(type=float, default=None)

    def __config_post_init__(self, config_dict):
        if self.label:
            self.labels.append(self.label)

        self.max_num_workers = max(2, self.num_workers)
        self.USER_ATTRIBS = []

        valid_disk_types = ["local-ssd", "pd-ssd", "pd-standard"]

        def format_disk_type(simple_type):
            return WORKER_DISK_TYPE_URL.format(
                project=self.project,
                region=self.region,
                disk_type=simple_type,
            )

        # worker_disk_type may or may not already be formatted as a URL
        if self.worker_disk_type is not None:
            if self.worker_disk_type in valid_disk_types:
                self.worker_disk_type = format_disk_type(self.worker_disk_type)
            elif self.worker_disk_type != format_disk_type(
                self.worker_disk_type.split("/")[-1]
            ):
                raise ValueError(
                    "Invalid pipeline_options.worker_disk_type: '{}'".format(
                        self.worker_disk_type
                    )
                )

        declared_config = self._as_dict()
        for key, value in config_dict.items():
            if key not in declared_config:
                # set user attributes to job_config - but only the first
                # level, for example, from this
                # pipeline_options:
                #    foo:
                #      key1: value1
                #      list1:
                #        - one
                #        - two
                # gets parsed to:
                # pipeline_options.foo ->
                #   {"key1": "value1", "alist": ["one", "two"]}
                setattr(self, key, value)
                # key track of user-set attributes so that when as_dict()
                # is called, we re-add it to what _as_dict() returns
                self.USER_ATTRIBS.append({key: value})

    def _as_dict(self):
        return attr.asdict(
            self, filter=lambda x, _: x.name not in self.ATTRIBS_TO_SKIP
        )

    def as_dict(self):
        """Return a dictionary representation of the
        :class:`KlioPipelineConfig` object.
        """
        config_dict = self._as_dict()
        for attrib in self.USER_ATTRIBS:
            for key, value in attrib.items():
                config_dict[key] = value

        return config_dict

    def __repr__(self):
        return "KlioPipelineConfig(job_name=%s)" % self.job_name



# Copyright 2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import json

import pytest
import yaml

from klio_core import exceptions
from klio_core.config import _preprocessing


@pytest.fixture
def kcp():
    return _preprocessing.KlioConfigPreprocessor


@pytest.fixture
def job_raw_config():
    return {
        "allow_non_klio_messages": False,
        "events": {
            "inputs": {
                "file0": {
                    "type": "file",
                    "location": "gs://test/yesterday.txt",
                },
                "file1": {"type": "file", "location": "gs://test/today.txt"},
            }
        },
    }


@pytest.fixture
def job_raw_config_templated():
    return {
        "events": {
            "inputs": {
                "bigquery0": {
                    "type": "bigquery",
                    "table": "testing.top_tracks_global",
                    "partition": "$YESTERDAY",
                },
                "bigquery1": {
                    "type": "bigquery",
                    "table": "testing.top_artists_global",
                    "partition": "$TODAY",
                },
                "bigquery2": {
                    "type": "bigquery",
                    "table": "testing.${GENRE}_join_$COUNTRY",
                    "partition": "$TODAY",
                },
            }
        }
    }


@pytest.mark.parametrize(
    "overrides_dict,expected",
    (
        # No overrides given - no changes in returned dict
        (
            {},
            {
                "allow_non_klio_messages": False,
                "events": {
                    "inputs": {
                        "file0": {
                            "type": "file",
                            "location": "gs://test/yesterday.txt",
                        },
                        "file1": {
                            "type": "file",
                            "location": "gs://test/today.txt",
                        },
                    }
                },
            },
        ),
        # Override dict given but no changes - no changes in returned dict
        (
            {"allow_non_klio_messages": True},
            {
                "allow_non_klio_messages": True,
                "events": {
                    "inputs": {
                        "file0": {
                            "type": "file",
                            "location": "gs://test/yesterday.txt",
                        },
                        "file1": {
                            "type": "file",
                            "location": "gs://test/today.txt",
                        },
                    }
                },
            },
        ),
        # Override current values
        (
            {
                "events.inputs.file0.location": "gs://test/12-31-2019.txt",
                "events.inputs.file1.location": "gs://test/01-01-2020.txt",
            },
            {
                "allow_non_klio_messages": False,
                "events": {
                    "inputs": {
                        "file0": {
                            "type": "file",
                            "location": "gs://test/12-31-2019.txt",
                        },
                        "file1": {
                            "type": "file",
                            "location": "gs://test/01-01-2020.txt",
                        },
                    }
                },
            },
        ),
        # Add new values
        (
            {
                "events.inputs.file2.location": "gs://test/01-01-2020.txt",
                "events.inputs.file2.type": "file",
            },
            {
                "allow_non_klio_messages": False,
                "events": {
                    "inputs": {
                        "file0": {
                            "type": "file",
                            "location": "gs://test/yesterday.txt",
                        },
                        "file1": {
                            "type": "file",
                            "location": "gs://test/today.txt",
                        },
                        "file2": {
                            "type": "file",
                            "location": "gs://test/01-01-2020.txt",
                        },
                    }
                },
            },
        ),
    ),
)
def test_apply_overrides(kcp, job_raw_config, overrides_dict, expected):
    # One key override
    new_dict = kcp._apply_overrides(job_raw_config, overrides_dict)
    assert expected == new_dict


@pytest.mark.parametrize(
    "template_dict,expected",
    (
        # One or more missing templates should raise a KeyError
        (
            {
                "YESTERDAY": "12-31-2019",
                "TODAY": "01-01-2020",
                "COUNTRY": "sto",
            },
            None,
        ),
        (
            {
                "YESTERDAY": "12-31-2019",
                "TODAY": "01-01-2020",
                "GENRE": "rock",
                "COUNTRY": "sto",
            },
            {
                "events": {
                    "inputs": {
                        "bigquery0": {
                            "type": "bigquery",
                            "table": "testing.top_tracks_global",
                            "partition": "12-31-2019",
                        },
                        "bigquery1": {
                            "type": "bigquery",
                            "table": "testing.top_artists_global",
                            "partition": "01-01-2020",
                        },
                        "bigquery2": {
                            "type": "bigquery",
                            "table": "testing.rock_join_sto",
                            "partition": "01-01-2020",
                        },
                    }
                }
            },
        ),
    ),
)
def test_apply_templates(
    kcp, job_raw_config_templated, template_dict, expected
):
    raw_config_str = json.dumps(job_raw_config_templated)
    if not expected:
        with pytest.raises(
            exceptions.KlioConfigTemplatingException, match=r"'GENRE' .*"
        ):
            kcp._apply_templates(raw_config_str, template_dict)
    else:
        new_dict = kcp._apply_templates(raw_config_str, template_dict)
        assert isinstance(new_dict, str)
        assert expected == json.loads(new_dict)


def test_transform_io(kcp):
    config = {
        "job_config": {
            "events": {
                "inputs": [
                    {"type": "bq", "key": "value"},
                    {"type": "bq", "key": "value2"},
                    {"type": "gcs", "gcskey": "gcsvalue"},
                ],
                "outputs": [
                    {"type": "bq", "key": "value", "name": "mybq"},
                    {"type": "bq", "key": "value2"},
                ],
            },
            "data": {
                "inputs": [],
                "outputs": [
                    {"type": "bq", "key": "value", "name": "mybq"},
                    {"type": "bq", "key": "value2"},
                ],
            },
        }
    }
    expected = {
        "job_config": {
            "events": {
                "inputs": {
                    "bq0": {"type": "bq", "key": "value"},
                    "bq1": {"type": "bq", "key": "value2"},
                    "gcs0": {"type": "gcs", "gcskey": "gcsvalue"},
                },
                "outputs": {
                    "mybq": {"type": "bq", "key": "value"},
                    "bq0": {"type": "bq", "key": "value2"},
                },
            },
            "data": {
                "inputs": {},
                "outputs": {
                    "mybq": {"type": "bq", "key": "value"},
                    "bq0": {"type": "bq", "key": "value2"},
                },
            },
        }
    }
    actual = kcp._transform_io_sections(config)

    assert actual == expected


def test_apply_plugin_preprocessing(mocker, kcp):

    config = {
        "version": 1,
        "job_name": "test_job",
        "job_config": {},
    }
    raw_config = yaml.dump(config)

    def add_field(config_dict):
        config_dict["version"] = 2
        return config_dict

    expected_config = config.copy()
    expected_config["version"] = 2

    mocker.patch.object(kcp, "PLUGIN_PREPROCESSORS", [add_field])

    processed_config = kcp.process(raw_config, [], [])

    assert expected_config == processed_config



# Copyright 2019-2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Shared variables for use within the Klio ecosystem."""


# From https://cloud.google.com/dataflow/docs/concepts/regional-endpoints
# Last pulled: 2020-10-01
# TODO: Figure out a way to get this dynamically (and maybe fall back to
#       hard-coded) so we don't have to manually keep this up to date.
DATAFLOW_REGIONS = (
    "asia-east1",
    "asia-northeast1",
    "asia-southeast1",
    "australia-southeast1",
    "europe-west1",
    "europe-west2",
    "europe-west3",
    "europe-west4",
    "northamerica-northeast1",
    "us-central1",
    "us-east1",
    "us-east4",
    "us-west1",
)
"""Default tuple of regions/locations for which to query."""



# Copyright 2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import logging

import attr
import yaml

from klio_core.config import _converters as converters

logger = logging.getLogger("klio")


def field(**kwargs):
    """Convienence function to create an attribute with parameters tuned for
    config objects.

    """
    # fill in some useful defaults
    if "default" not in kwargs:
        kwargs["default"] = converters.UNSET_REQUIRED_VALUE
    if "repr" not in kwargs:
        kwargs["repr"] = False

    return attr.attrib(**kwargs)


@attr.attrs
class WrappedValidator(object):
    """A Simple validator that just wraps another validator and prepends the
    full config key to any error
    """

    key = attr.attrib()
    inner_validator = attr.attrib()

    def __call__(self, cls, attrib, value):
        try:
            self.inner_validator(cls, attrib, value)
        except Exception as e:
            raise Exception("{}: {}".format(self.key, e))


class config_object(object):
    """Class used as a decorator to add some specialized functionality to
    classes that parse config data.  It is intended to be used with attrs and
    adds settings to any detected attributes defined with `attr.attrib()`.

    The class' __init__ method generated by attrs is augmented to accept a
    special `config_dict` parameter which will do some pre-processing and then
    initialize the class

    Any attrib without a default set will have it set to UNSET_REQUIRED_VALUE,
    which is then detected in the converter that's added to raise a properly
    formatted error message when a value isn't provided by the user.

    If the attrib specifies a recognized `type` (see Converters), a converter
    for that type will be added to ensure any provided values are of that type
    (or can be easily coerced to that type).

    The decorated class may define a `__config_post_init__(self, config_dict)`
    method which will be called after initialization (and after attrs own
    `__attrs_post_init__`), where `config_dict` is the original unmodified
    dict.

    Args:
        key_prefix (str): prefix prepended to any key names in error messages

        allow_unknown_keys (bool): if True, any keys in `config_dict` that are
            not registered `attrib()`s will be ignored and not raise an unknown
            key error.

        ignore_keys (list): list of keys to explicitly ignore (only needed when
            `allow_unknown_keys` si `True`

    """

    def __init__(
        self, key_prefix=None, allow_unknown_keys=False, ignore_keys=None
    ):
        self.key_prefix = key_prefix
        self.allow_unknown_keys = allow_unknown_keys
        self.ignore_keys = [] if ignore_keys is None else ignore_keys

    def _full_key(self, key):
        if self.key_prefix is None:
            return key
        return self.key_prefix + "." + key

    def _prepare_config_dict(self, known_keys, config_dict, **extra_kwargs):
        if self.allow_unknown_keys:
            actual_keys = extra_kwargs
            for key in known_keys:
                if key in config_dict:
                    actual_keys[key] = config_dict[key]
        else:
            actual_keys = config_dict.copy()
            for key, value in extra_kwargs.items():
                actual_keys[key] = extra_kwargs[key]

            for key, value in actual_keys.items():
                if key not in known_keys:
                    raise Exception(
                        "Unexpected config key {}".format(self._full_key(key))
                    )

        for key in self.ignore_keys:
            actual_keys.pop(key, None)

        return actual_keys

    def __call__(self, cls):

        known_keys = []
        # add type checking converters to attribs before running through
        # attr.attrs()
        for key, attrib in cls.__dict__.items():
            if hasattr(attrib, "_default") and hasattr(attrib, "converter"):
                if attrib._default == attr.NOTHING:
                    attrib._default = converters.UNSET_REQUIRED_VALUE

                # wrap any validator in one that includes the key in the error
                # message
                if attrib._validator is not None:
                    attrib._validator = WrappedValidator(
                        self._full_key(key), attrib._validator
                    )

                if attrib.type is not None:
                    attrib.converter = converters.Converters.for_type(
                        attrib.type, self._full_key(key)
                    )
                else:
                    attrib.converter = converters.ConfigValueConverter(
                        self._full_key(key)
                    ).validate

                known_keys.append(key)

        # now we can let attrs work its magic
        attrib_cls = attr.attrs(cls)

        # lastly we'll replace the __init__ provided by attrs (which accepts
        # arguments directly) with one that accepts a dict along with
        # additional kargs, merges and cleans it up before passing it to attrs

        original_init = attrib_cls.__init__

        # need to alias self so we don't confuse the decorated class with this
        # decorator class!
        me = self

        def init_from_dict(self, config_dict={}, **extra_kwargs):
            original_init(
                self,
                **me._prepare_config_dict(
                    known_keys, config_dict, **extra_kwargs
                )
            )
            if hasattr(self, "__config_post_init__"):
                self.__config_post_init__(config_dict)

        attrib_cls.__init__ = init_from_dict
        attrib_cls.from_values = original_init
        return attrib_cls


class IndentListDumper(yaml.Dumper):
    """Force indentation for lists for better visual understanding.

    Instead of this:

        foo:
          bar:
          - one
          - two
          - three

    Format list indentations like this:
        foo:
          bar:
            - one
            - two
            - three
    """

    def increase_indent(self, flow=False, indentless=False):
        return super(IndentListDumper, self).increase_indent(flow, False)



# Copyright 2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import logging


logger = logging.getLogger("klio")


class _UnsetRequiredValue(object):
    """
    Sentinel class used on as the default for attributes that require a value.
    ``_UnsetRequiredValue`` is a singleton. There is only ever one of it.
    """

    _singleton = None

    def __new__(cls):
        if _UnsetRequiredValue._singleton is None:
            _UnsetRequiredValue._singleton = super(
                _UnsetRequiredValue, cls
            ).__new__(cls)
        return _UnsetRequiredValue._singleton

    def __repr__(self):
        return "UNSET_REQUIRED_VALUE"


UNSET_REQUIRED_VALUE = _UnsetRequiredValue()
"""
This is set as the default for an attribute when we want no default.  This way
attrs itself doesn't raise an exception and instead the value is passed into
our converter which will raise a properly formatted error
"""


class ConfigValueConverter(object):
    """Base class for converters automatically added to any `attrs` `attrib`
    created in a class decorated with `config_object`.  These converters
    properly handle raising Exceptions when a required value is not provided,
    as well as more strict type checking and type coercion.

    """

    def __init__(self, key):
        self.key = key

    def validate(self, value):
        if value == UNSET_REQUIRED_VALUE:
            raise Exception("missing required key: {}".format(self.key))
        elif value is None:
            return value
        return self._validate_value(value)

    def _validate_value(self, value):
        return value


class TypeCheckingConverter(ConfigValueConverter):
    def __init__(self, key, type):
        super().__init__(key)
        self.type = type

    def _wrong_type_msg(self, wrong_value):
        return (
            "{}: expected value of type '{}'" ", got a '{}' value instead"
        ).format(self.key, self.type.__name__, type(wrong_value).__name__)


class SingleValueConverter(TypeCheckingConverter):
    def validate(self, value):
        if isinstance(value, dict) or isinstance(value, list):
            self._wrong_type_message(value)

        return super().validate(value)


class IntConverter(SingleValueConverter):
    def __init__(self, key_prefix):
        super().__init__(key_prefix, int)

    def _validate_value(self, value):
        try:
            return int(value)
        except Exception:
            raise Exception(
                "{}: expected numeric value, got '{}'".format(self.key, value)
            )


class BoolConverter(SingleValueConverter):
    def __init__(self, key_prefix):
        super().__init__(key_prefix, bool)

    def _validate_value(self, value):
        if value is None:
            return value
        try:
            new_value = bool(value)
            if not isinstance(value, bool):
                logger.warning(
                    (
                        "{}: found non-bool value '{}'"
                        ", converting to bool '{}'"
                    ).format(self.key, value, new_value)
                )
            return new_value
        except Exception:
            raise Exception(
                "{}: expected boolean value, got '{}'".format(self.key, value)
            )


class StringConverter(SingleValueConverter):
    def __init__(self, key_prefix):
        super().__init__(key_prefix, str)

    def _validate_value(self, value):
        if value is None:
            return value

        if not isinstance(value, str):
            logger.warning(
                "{}: found non-string value '{}', converting to str".format(
                    self.key, value
                )
            )
        return str(value)


class Converters(object):

    TYPES = {
        int: IntConverter,
        str: StringConverter,
        bool: BoolConverter,
    }

    @classmethod
    def for_type(cls, type, key):
        if type in cls.TYPES:
            return cls.TYPES[type](key).validate
        else:
            return ConfigValueConverter(key).validate



# Copyright 2019-2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os

from unittest import mock

import pytest

from google.api_core import exceptions as gapi_exceptions

from klio_core import utils


def test_set_global():
    utils.set_global("set-a-value", "a-value")

    actual = getattr(utils, "klio_global_state_set-a-value", None)
    assert "a-value" == actual

    delattr(utils, "klio_global_state_set-a-value")


def test_get_global():
    setattr(utils, "klio_global_state_get-a-value", "a-value")

    actual = utils.get_global("get-a-value")
    assert "a-value" == actual

    delattr(utils, "klio_global_state_get-a-value")


@pytest.mark.parametrize("exists", (True, False))
def test_delete_global(exists):
    if exists:
        setattr(utils, "klio_global_state_delete-a-value", "a-value")

    utils.delete_global("delete-a-value")

    actual = getattr(utils, "klio_global_state_delete-a-value", None)

    assert not actual


@pytest.mark.parametrize(
    "set_value,callable_init", ((True, False), (False, True), (False, False))
)
def test_get_or_initialize_global(set_value, callable_init, mocker):
    if set_value:
        setattr(utils, "klio_global_state_get-or-init-value", "a-value")
    if callable_init:
        initializer = mocker.Mock(return_value="a-value")
    else:
        initializer = "a-value"

    actual = utils.get_or_initialize_global("get-or-init-value", initializer)

    assert "a-value" == actual
    if not isinstance(initializer, str):
        initializer.assert_called_once_with()

    delattr(utils, "klio_global_state_get-or-init-value")


@pytest.fixture
def mock_publisher(mocker, monkeypatch):
    mock = mocker.Mock()
    monkeypatch.setattr(utils.pubsub, "PublisherClient", mock)
    return mock


def test_private_get_publisher(mock_publisher):
    ret_publisher = utils._get_publisher("a-topic")

    mock_publisher.assert_called_once_with()
    mock_publisher.return_value.create_topic.assert_called_once_with("a-topic")

    assert mock_publisher.return_value == ret_publisher


def test_private_get_publisher_topic_exists(mock_publisher):
    client = mock_publisher.return_value
    client.create_topic.side_effect = gapi_exceptions.AlreadyExists("foo")

    ret_publisher = utils._get_publisher("a-topic")

    mock_publisher.assert_called_once_with()
    client.create_topic.assert_called_once_with("a-topic")

    assert client == ret_publisher


def test_private_get_publisher_raises(mock_publisher):
    client = mock_publisher.return_value
    client.create_topic.side_effect = Exception("foo")

    with pytest.raises(Exception, match="foo"):
        utils._get_publisher("a-topic")

    mock_publisher.assert_called_once_with()
    client.create_topic.assert_called_once_with("a-topic")


@pytest.mark.parametrize("in_globals", (True, False))
def test_get_publisher(in_globals, mock_publisher):
    client = mock_publisher.return_value
    if in_globals:
        setattr(utils, "klio_global_state_publisher_a-topic", client)

    ret_publisher = utils.get_publisher("a-topic")

    assert client == ret_publisher
    delattr(utils, "klio_global_state_publisher_a-topic")


#########################
# Config Utils tests
#########################
@pytest.fixture
def patch_os_getcwd(monkeypatch, tmpdir):
    test_dir = str(tmpdir.mkdir("testing"))
    monkeypatch.setattr(os, "getcwd", lambda: test_dir)
    return test_dir


def test_get_config_by_path(mocker, monkeypatch):
    m_open = mocker.mock_open()
    mock_open = mocker.patch("klio_core.utils.open", m_open)

    mock_safe_load = mocker.Mock()
    monkeypatch.setattr(utils.yaml, "safe_load", mock_safe_load)

    path = "path/to/a/file"
    utils.get_config_by_path(path)

    mock_open.assert_called_once_with(path)


def test_get_config_by_path_error(mocker, monkeypatch, caplog):
    m_open = mocker.mock_open()
    mock_open = mocker.patch("klio_core.utils.open", m_open)
    mock_open.side_effect = IOError

    with pytest.raises(SystemExit):
        path = "path/to/a/file"
        utils.get_config_by_path(path)

    assert 1 == len(caplog.records)


#########################
# Cli/exec command tests
#########################
@pytest.mark.parametrize(
    "image",
    (
        "dataflow.gcr.io/v1beta3/python",
        "dataflow.gcr.io/v1beta3/python-base",
        "dataflow.gcr.io/v1beta3/python-fnapi",
    ),
)
def test_warn_if_py2_job(image, patch_os_getcwd, mocker):
    dockerfile = (
        '## -*- docker-image-name: "gcr.io/foo/bar" -*-\n'
        "FROM {image}:1.2.3\n"
        'LABEL maintainer "me@example.com"\n'
    ).format(image=image)

    m_open = mock.mock_open(read_data=dockerfile)
    mock_open = mocker.patch("klio_core.utils.open", m_open)
    m_is_file = mocker.Mock()
    m_is_file.return_value = True
    mock_is_file = mocker.patch("klio_core.utils.os.path.isfile", m_is_file)

    warn_msg = (
        "Python 2 support in Klio is deprecated. "
        "Please upgrade to Python 3.5+"
    )
    with pytest.warns(UserWarning, match=warn_msg):
        utils.warn_if_py2_job(patch_os_getcwd)

    exp_read_file = os.path.join(patch_os_getcwd, "Dockerfile")
    mock_is_file.assert_called_once_with(exp_read_file)
    mock_open.assert_called_once_with(exp_read_file, "r")


@pytest.mark.parametrize("has_from_line", (True, False))
@pytest.mark.parametrize("file_exists", (True, False))
def test_warn_if_py2_job_no_warn(
    has_from_line, file_exists, patch_os_getcwd, mocker
):
    from_line = "\n"
    if has_from_line:
        from_line = "FROM dataflow.gcr.io/v1beta3/python36-fnapi:1.2.3\n"

    dockerfile = (
        '## -*- docker-image-name: "gcr.io/foo/bar" -*-\n'
        + from_line
        + 'LABEL maintainer "me@example.com"\n'
    )

    m_open = mock.mock_open(read_data=dockerfile)
    mock_open = mocker.patch("klio_core.utils.open", m_open)
    m_is_file = mocker.Mock()
    m_is_file.return_value = file_exists
    mock_is_file = mocker.patch("klio_core.utils.os.path.isfile", m_is_file)

    utils.warn_if_py2_job(patch_os_getcwd)

    exp_read_file = os.path.join(patch_os_getcwd, "Dockerfile")
    mock_is_file.assert_called_once_with(exp_read_file)
    if file_exists:
        mock_open.assert_called_once_with(exp_read_file, "r")


@pytest.mark.parametrize(
    "job_dir,conf_file",
    (
        (None, None),
        (None, "klio-job2.yaml"),
        ("foo/bar", None),
        ("foo/bar", "klio-job2.yaml"),
    ),
)
def test_get_config_job_dir(job_dir, conf_file, patch_os_getcwd):
    exp_job_dir = patch_os_getcwd
    if job_dir:
        exp_job_dir = os.path.abspath(os.path.join(patch_os_getcwd, job_dir))
    exp_conf_file = conf_file or os.path.join(exp_job_dir, "klio-job.yaml")
    if job_dir and conf_file:
        exp_conf_file = os.path.join(job_dir, conf_file)

    ret_job_dir, ret_conf_file = utils.get_config_job_dir(job_dir, conf_file)

    assert exp_job_dir == ret_job_dir
    assert exp_conf_file == ret_conf_file



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: klio.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='klio.proto',
  package='',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\nklio.proto\x1a\x1fgoogle/protobuf/timestamp.proto\"\xed\x04\n\x0bKlioMessage\x12\'\n\x08metadata\x18\x01 \x01(\x0b\x32\x15.KlioMessage.Metadata\x12\x1f\n\x04\x64\x61ta\x18\x02 \x01(\x0b\x32\x11.KlioMessage.Data\x12\x19\n\x07version\x18\x03 \x01(\x0e\x32\x08.Version\x1a\xbb\x03\n\x08Metadata\x12\x1c\n\ndownstream\x18\x01 \x03(\x0b\x32\x08.KlioJob\x12\x19\n\x07visited\x18\x02 \x03(\x0b\x32\x08.KlioJob\x12+\n\rjob_audit_log\x18\x03 \x03(\x0b\x32\x14.KlioJobAuditLogItem\x12\r\n\x05\x66orce\x18\x04 \x01(\x08\x12\x0c\n\x04ping\x18\x05 \x01(\x08\x12=\n\x13intended_recipients\x18\x06 \x01(\x0b\x32 .KlioMessage.Metadata.Recipients\x1a\xec\x01\n\nRecipients\x12\x39\n\x06\x61nyone\x18\x01 \x01(\x0b\x32\'.KlioMessage.Metadata.Recipients.AnyoneH\x00\x12;\n\x07limited\x18\x02 \x01(\x0b\x32(.KlioMessage.Metadata.Recipients.LimitedH\x00\x1a\x08\n\x06\x41nyone\x1aN\n\x07Limited\x12\x1c\n\nrecipients\x18\x01 \x03(\x0b\x32\x08.KlioJob\x12%\n\x13trigger_children_of\x18\x02 \x01(\x0b\x32\x08.KlioJobB\x0c\n\nrecipients\x1a;\n\x04\x44\x61ta\x12\x11\n\tentity_id\x18\x01 \x01(\t\x12\x0f\n\x07payload\x18\x02 \x01(\x0c\x12\x0f\n\x07\x65lement\x18\x03 \x01(\x0c\"\x9b\x01\n\x07KlioJob\x12\x10\n\x08job_name\x18\x01 \x01(\t\x12\x13\n\x0bgcp_project\x18\x02 \x01(\t\x12!\n\x06inputs\x18\x03 \x03(\x0b\x32\x11.KlioJob.JobInput\x1a\x46\n\x08JobInput\x12\r\n\x05topic\x18\x01 \x01(\t\x12\x14\n\x0csubscription\x18\x02 \x01(\t\x12\x15\n\rdata_location\x18\x03 \x01(\t\"`\n\x13KlioJobAuditLogItem\x12-\n\ttimestamp\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x1a\n\x08klio_job\x18\x02 \x01(\x0b\x32\x08.KlioJob*&\n\x07Version\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x06\n\x02V1\x10\x01\x12\x06\n\x02V2\x10\x02\x62\x06proto3'
  ,
  dependencies=[google_dot_protobuf_dot_timestamp__pb2.DESCRIPTOR,])

_VERSION = _descriptor.EnumDescriptor(
  name='Version',
  full_name='Version',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNKNOWN', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='V1', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='V2', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=927,
  serialized_end=965,
)
_sym_db.RegisterEnumDescriptor(_VERSION)

Version = enum_type_wrapper.EnumTypeWrapper(_VERSION)
UNKNOWN = 0
V1 = 1
V2 = 2



_KLIOMESSAGE_METADATA_RECIPIENTS_ANYONE = _descriptor.Descriptor(
  name='Anyone',
  full_name='KlioMessage.Metadata.Recipients.Anyone',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=506,
  serialized_end=514,
)

_KLIOMESSAGE_METADATA_RECIPIENTS_LIMITED = _descriptor.Descriptor(
  name='Limited',
  full_name='KlioMessage.Metadata.Recipients.Limited',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='recipients', full_name='KlioMessage.Metadata.Recipients.Limited.recipients', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='trigger_children_of', full_name='KlioMessage.Metadata.Recipients.Limited.trigger_children_of', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=516,
  serialized_end=594,
)

_KLIOMESSAGE_METADATA_RECIPIENTS = _descriptor.Descriptor(
  name='Recipients',
  full_name='KlioMessage.Metadata.Recipients',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='anyone', full_name='KlioMessage.Metadata.Recipients.anyone', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='limited', full_name='KlioMessage.Metadata.Recipients.limited', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_KLIOMESSAGE_METADATA_RECIPIENTS_ANYONE, _KLIOMESSAGE_METADATA_RECIPIENTS_LIMITED, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='recipients', full_name='KlioMessage.Metadata.Recipients.recipients',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=372,
  serialized_end=608,
)

_KLIOMESSAGE_METADATA = _descriptor.Descriptor(
  name='Metadata',
  full_name='KlioMessage.Metadata',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='downstream', full_name='KlioMessage.Metadata.downstream', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='visited', full_name='KlioMessage.Metadata.visited', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='job_audit_log', full_name='KlioMessage.Metadata.job_audit_log', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='force', full_name='KlioMessage.Metadata.force', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ping', full_name='KlioMessage.Metadata.ping', index=4,
      number=5, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='intended_recipients', full_name='KlioMessage.Metadata.intended_recipients', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_KLIOMESSAGE_METADATA_RECIPIENTS, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=165,
  serialized_end=608,
)

_KLIOMESSAGE_DATA = _descriptor.Descriptor(
  name='Data',
  full_name='KlioMessage.Data',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='entity_id', full_name='KlioMessage.Data.entity_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='payload', full_name='KlioMessage.Data.payload', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='element', full_name='KlioMessage.Data.element', index=2,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=610,
  serialized_end=669,
)

_KLIOMESSAGE = _descriptor.Descriptor(
  name='KlioMessage',
  full_name='KlioMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='metadata', full_name='KlioMessage.metadata', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='data', full_name='KlioMessage.data', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='version', full_name='KlioMessage.version', index=2,
      number=3, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_KLIOMESSAGE_METADATA, _KLIOMESSAGE_DATA, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=48,
  serialized_end=669,
)


_KLIOJOB_JOBINPUT = _descriptor.Descriptor(
  name='JobInput',
  full_name='KlioJob.JobInput',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='topic', full_name='KlioJob.JobInput.topic', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='subscription', full_name='KlioJob.JobInput.subscription', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='data_location', full_name='KlioJob.JobInput.data_location', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=757,
  serialized_end=827,
)

_KLIOJOB = _descriptor.Descriptor(
  name='KlioJob',
  full_name='KlioJob',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='job_name', full_name='KlioJob.job_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='gcp_project', full_name='KlioJob.gcp_project', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='inputs', full_name='KlioJob.inputs', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_KLIOJOB_JOBINPUT, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=672,
  serialized_end=827,
)


_KLIOJOBAUDITLOGITEM = _descriptor.Descriptor(
  name='KlioJobAuditLogItem',
  full_name='KlioJobAuditLogItem',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='timestamp', full_name='KlioJobAuditLogItem.timestamp', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='klio_job', full_name='KlioJobAuditLogItem.klio_job', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=829,
  serialized_end=925,
)

_KLIOMESSAGE_METADATA_RECIPIENTS_ANYONE.containing_type = _KLIOMESSAGE_METADATA_RECIPIENTS
_KLIOMESSAGE_METADATA_RECIPIENTS_LIMITED.fields_by_name['recipients'].message_type = _KLIOJOB
_KLIOMESSAGE_METADATA_RECIPIENTS_LIMITED.fields_by_name['trigger_children_of'].message_type = _KLIOJOB
_KLIOMESSAGE_METADATA_RECIPIENTS_LIMITED.containing_type = _KLIOMESSAGE_METADATA_RECIPIENTS
_KLIOMESSAGE_METADATA_RECIPIENTS.fields_by_name['anyone'].message_type = _KLIOMESSAGE_METADATA_RECIPIENTS_ANYONE
_KLIOMESSAGE_METADATA_RECIPIENTS.fields_by_name['limited'].message_type = _KLIOMESSAGE_METADATA_RECIPIENTS_LIMITED
_KLIOMESSAGE_METADATA_RECIPIENTS.containing_type = _KLIOMESSAGE_METADATA
_KLIOMESSAGE_METADATA_RECIPIENTS.oneofs_by_name['recipients'].fields.append(
  _KLIOMESSAGE_METADATA_RECIPIENTS.fields_by_name['anyone'])
_KLIOMESSAGE_METADATA_RECIPIENTS.fields_by_name['anyone'].containing_oneof = _KLIOMESSAGE_METADATA_RECIPIENTS.oneofs_by_name['recipients']
_KLIOMESSAGE_METADATA_RECIPIENTS.oneofs_by_name['recipients'].fields.append(
  _KLIOMESSAGE_METADATA_RECIPIENTS.fields_by_name['limited'])
_KLIOMESSAGE_METADATA_RECIPIENTS.fields_by_name['limited'].containing_oneof = _KLIOMESSAGE_METADATA_RECIPIENTS.oneofs_by_name['recipients']
_KLIOMESSAGE_METADATA.fields_by_name['downstream'].message_type = _KLIOJOB
_KLIOMESSAGE_METADATA.fields_by_name['visited'].message_type = _KLIOJOB
_KLIOMESSAGE_METADATA.fields_by_name['job_audit_log'].message_type = _KLIOJOBAUDITLOGITEM
_KLIOMESSAGE_METADATA.fields_by_name['intended_recipients'].message_type = _KLIOMESSAGE_METADATA_RECIPIENTS
_KLIOMESSAGE_METADATA.containing_type = _KLIOMESSAGE
_KLIOMESSAGE_DATA.containing_type = _KLIOMESSAGE
_KLIOMESSAGE.fields_by_name['metadata'].message_type = _KLIOMESSAGE_METADATA
_KLIOMESSAGE.fields_by_name['data'].message_type = _KLIOMESSAGE_DATA
_KLIOMESSAGE.fields_by_name['version'].enum_type = _VERSION
_KLIOJOB_JOBINPUT.containing_type = _KLIOJOB
_KLIOJOB.fields_by_name['inputs'].message_type = _KLIOJOB_JOBINPUT
_KLIOJOBAUDITLOGITEM.fields_by_name['timestamp'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_KLIOJOBAUDITLOGITEM.fields_by_name['klio_job'].message_type = _KLIOJOB
DESCRIPTOR.message_types_by_name['KlioMessage'] = _KLIOMESSAGE
DESCRIPTOR.message_types_by_name['KlioJob'] = _KLIOJOB
DESCRIPTOR.message_types_by_name['KlioJobAuditLogItem'] = _KLIOJOBAUDITLOGITEM
DESCRIPTOR.enum_types_by_name['Version'] = _VERSION
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

KlioMessage = _reflection.GeneratedProtocolMessageType('KlioMessage', (_message.Message,), {

  'Metadata' : _reflection.GeneratedProtocolMessageType('Metadata', (_message.Message,), {

    'Recipients' : _reflection.GeneratedProtocolMessageType('Recipients', (_message.Message,), {

      'Anyone' : _reflection.GeneratedProtocolMessageType('Anyone', (_message.Message,), {
        'DESCRIPTOR' : _KLIOMESSAGE_METADATA_RECIPIENTS_ANYONE,
        '__module__' : 'klio_pb2'
        # @@protoc_insertion_point(class_scope:KlioMessage.Metadata.Recipients.Anyone)
        })
      ,

      'Limited' : _reflection.GeneratedProtocolMessageType('Limited', (_message.Message,), {
        'DESCRIPTOR' : _KLIOMESSAGE_METADATA_RECIPIENTS_LIMITED,
        '__module__' : 'klio_pb2'
        # @@protoc_insertion_point(class_scope:KlioMessage.Metadata.Recipients.Limited)
        })
      ,
      'DESCRIPTOR' : _KLIOMESSAGE_METADATA_RECIPIENTS,
      '__module__' : 'klio_pb2'
      # @@protoc_insertion_point(class_scope:KlioMessage.Metadata.Recipients)
      })
    ,
    'DESCRIPTOR' : _KLIOMESSAGE_METADATA,
    '__module__' : 'klio_pb2'
    # @@protoc_insertion_point(class_scope:KlioMessage.Metadata)
    })
  ,

  'Data' : _reflection.GeneratedProtocolMessageType('Data', (_message.Message,), {
    'DESCRIPTOR' : _KLIOMESSAGE_DATA,
    '__module__' : 'klio_pb2'
    # @@protoc_insertion_point(class_scope:KlioMessage.Data)
    })
  ,
  'DESCRIPTOR' : _KLIOMESSAGE,
  '__module__' : 'klio_pb2'
  # @@protoc_insertion_point(class_scope:KlioMessage)
  })
_sym_db.RegisterMessage(KlioMessage)
_sym_db.RegisterMessage(KlioMessage.Metadata)
_sym_db.RegisterMessage(KlioMessage.Metadata.Recipients)
_sym_db.RegisterMessage(KlioMessage.Metadata.Recipients.Anyone)
_sym_db.RegisterMessage(KlioMessage.Metadata.Recipients.Limited)
_sym_db.RegisterMessage(KlioMessage.Data)

KlioJob = _reflection.GeneratedProtocolMessageType('KlioJob', (_message.Message,), {

  'JobInput' : _reflection.GeneratedProtocolMessageType('JobInput', (_message.Message,), {
    'DESCRIPTOR' : _KLIOJOB_JOBINPUT,
    '__module__' : 'klio_pb2'
    # @@protoc_insertion_point(class_scope:KlioJob.JobInput)
    })
  ,
  'DESCRIPTOR' : _KLIOJOB,
  '__module__' : 'klio_pb2'
  # @@protoc_insertion_point(class_scope:KlioJob)
  })
_sym_db.RegisterMessage(KlioJob)
_sym_db.RegisterMessage(KlioJob.JobInput)

KlioJobAuditLogItem = _reflection.GeneratedProtocolMessageType('KlioJobAuditLogItem', (_message.Message,), {
  'DESCRIPTOR' : _KLIOJOBAUDITLOGITEM,
  '__module__' : 'klio_pb2'
  # @@protoc_insertion_point(class_scope:KlioJobAuditLogItem)
  })
_sym_db.RegisterMessage(KlioJobAuditLogItem)


# @@protoc_insertion_point(module_scope)



# Copyright 2019-2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Utility functions for use within the Klio ecosystem."""

import functools
import logging
import os
import warnings

import attr
import yaml

from google.api_core import exceptions as gapi_exceptions
from google.cloud import pubsub

from klio_core import config
from klio_core import options


def _name(name):
    return "klio_global_state_%s" % name


def set_global(name, value):
    """Set a variable in the global namespace.

    Args:
        name (str): name of global variable.
        value (Any): value of global variable.
    """
    globals()[_name(name)] = value


def get_global(name):
    """Get a variable from the global namespace.

    Args:
        name (str): name of global variable.
    Returns:
        Any: value of global variable, or ``None`` if not set.
    """
    return globals().get(_name(name), None)


def delete_global(name):
    """Delete a variable from the global namespace.

    Args:
        name (str): name of global variable.
    """
    if _name(name) in globals():
        del globals()[_name(name)]


def get_or_initialize_global(name, initializer):
    """Get a global variable, initializing if does not exist.

    .. caution::

        Global variables may cause problems for jobs in Dataflow,
        particularly where jobs use more than one thread.

    Args:
        name (str): name of global variable.
        initializer (Any): initial value if ``name`` does not exist in the
            global namespace.

    Returns:
        Any: Value of the global variable.
    """
    value = get_global(name)
    if value is not None:
        return value
    if callable(initializer):
        value = initializer()
    else:
        value = initializer
    set_global(name, value)
    return value


def _get_publisher(topic):
    publisher = pubsub.PublisherClient()
    try:
        publisher.create_topic(topic)
    except gapi_exceptions.AlreadyExists:
        pass

    except Exception:
        raise  # to be handled by caller

    return publisher


def get_publisher(topic):
    """Get a publisher client for a given topic.

    Will first check if there is an already initialized client in the
    global namespace. Otherwise, initialize one then set it in the
    global namespace to avoid redundant initialization.

    Args:
        topic (str): Pub/Sub topic for the client with which to be
            initialized.
    Returns:
        google.cloud.pubsub_v1.publisher.client.Client: an initialized
        client for publishing to Google Pub/Sub.
    """
    key = "publisher_{}".format(topic)
    initializer = functools.partial(_get_publisher, topic)
    return get_or_initialize_global(key, initializer)


# makeshift Enums for py2 support
# my_enum = enum("FOO", "BAR") -> 0=my_enum.FOO, 1=my_enum.BAR
def enum(*sequential, **named):
    enums = dict(zip(sequential, range(len(sequential))), **named)
    return type("Enum", (), enums)


#############################
# Config utils
#############################
def get_config_by_path(config_filepath, parse_yaml=True):
    """Read in the file given at specified config path

    Args:
        config_filepath (str): File path to klio config file
        parse_yaml (bool): Whether to parse the given file path
            as yaml
    Returns:
        python object of yaml config file or bytes read from file
    """
    try:
        with open(config_filepath) as f:
            if parse_yaml:
                return yaml.safe_load(f)
            else:
                return f.read()
    except IOError:
        logging.error("Could not read config file {0}".format(config_filepath))
        raise SystemExit(1)


def get_config_job_dir(job_dir, config_file):
    """Read in the file given at specified config path

    Args:
        job_dir (str): File directory path to klio config file
        config_file (str): File name of config file, if not provided
            ``klio-job.yaml`` will be used
    Returns:
        job_dir - Absolute path to config file directory
        config_file - Path to config file
    """
    if job_dir and config_file:
        config_file = os.path.join(job_dir, config_file)

    elif not job_dir:
        job_dir = os.getcwd()

    job_dir = os.path.abspath(job_dir)

    if not config_file:
        config_file = os.path.join(job_dir, "klio-job.yaml")

    return job_dir, config_file


@attr.attrs
class KlioConfigMeta(object):
    """
    Meta data about klio config
        job_dir: resolved directory of job
        config_path: resolved path to config file
        config_file: user-override of config file (may be None)
    """

    # resolved directory of job
    job_dir = attr.attrib()

    # resolved path to config file
    config_path = attr.attrib()

    # user-override of config file (may be None)
    config_file = attr.attrib()


def with_klio_config(func):
    """Decorator for commands to automatically handle a number of options
    regarding config, and provide a properly constructed KlioConfig as an
    argument named `klio_config`.

    Be aware this is a function wrapper and must come after any other
    decorators for click options, arguments, etc.
    """

    @options.override
    @options.template
    @options.job_dir
    @options.config_file
    def wrapper(*args, **kwargs):
        raw_overrides = kwargs.pop("override")
        raw_templates = kwargs.pop("template")
        job_dir = kwargs.pop("job_dir")
        config_file = kwargs.pop("config_file")
        job_dir, config_path = get_config_job_dir(job_dir, config_file)

        warn_if_py2_job(job_dir)

        raw_config_data = get_config_by_path(config_path)

        conf = config.KlioConfig(
            raw_config_data,
            raw_templates=raw_templates,
            raw_overrides=raw_overrides,
        )

        meta = KlioConfigMeta(
            job_dir=job_dir, config_file=config_file, config_path=config_path,
        )

        kwargs["klio_config"] = conf
        kwargs["config_meta"] = meta

        func(*args, **kwargs)

    return functools.update_wrapper(wrapper, func)


#############################
# Cli/exec utils
#############################
def warn_if_py2_job(job_dir):
    dockerfile_path = os.path.join(job_dir, "Dockerfile")
    if not os.path.isfile(dockerfile_path):
        return

    from_line = None
    with open(dockerfile_path, "r") as f:
        for line in f.readlines():
            if line.startswith("FROM"):
                from_line = line
                break
    if not from_line:
        # not having a FROM line will break elsewhere, so no need to take
        # care of it here
        return

    py2_dataflow_images = [
        "dataflow.gcr.io/v1beta3/python",
        "dataflow.gcr.io/v1beta3/python-base",
        "dataflow.gcr.io/v1beta3/python-fnapi",
    ]
    from_image = from_line.lstrip("FROM ").split(":")[0]
    if from_image in py2_dataflow_images:
        msg = (
            "Python 2 support in Klio is deprecated. Please upgrade "
            "to Python 3.5+."
        )
        warnings.warn(msg, category=UserWarning)



# Copyright 2019-2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import copy
import inspect

import dill
import pytest

from klio_core import config
from klio_core.config import _io as io


@pytest.fixture
def job_config_dict():
    return {
        "metrics": {"logger": {}},
        "events": {
            "inputs": {
                "pubsub0": {
                    "type": "pubsub",
                    "topic": "test-parent-job-out",
                    "subscription": "test-parent-job-out-sub",
                },
            },
            "outputs": {
                "pubsub0": {"type": "pubsub", "topic": "test-job-out"}
            },
        },
        "data": {
            "inputs": {
                "gcs0": {
                    "type": "GCS",
                    "location": "gs://sigint-output/test-parent-job-out",
                }
            },
            "outputs": {
                "gcs0": {
                    "type": "GCS",
                    "location": "gs://sigint-output/test-job-out",
                }
            },
        },
        "more": "config",
        "that": {"the": "user"},
        "might": ["include"],
        "blocking": False,
    }


@pytest.fixture
def final_job_config_dict():
    return {
        "metrics": {"logger": {}},
        "events": {
            "inputs": [
                {
                    "type": "pubsub",
                    "topic": "test-parent-job-out",
                    "subscription": "test-parent-job-out-sub",
                    "skip_klio_read": False,
                },
            ],
            "outputs": [
                {
                    "type": "pubsub",
                    "topic": "test-job-out",
                    "skip_klio_write": False,
                }
            ],
        },
        "data": {
            "inputs": [
                {
                    "type": "gcs",
                    "location": "gs://sigint-output/test-parent-job-out",
                    "skip_klio_existence_check": False,
                    "file_suffix": "",
                    "ping": False,
                }
            ],
            "outputs": [
                {
                    "type": "gcs",
                    "file_suffix": "",
                    "force": False,
                    "location": "gs://sigint-output/test-job-out",
                    "skip_klio_existence_check": False,
                }
            ],
        },
        "more": "config",
        "that": {"the": "user"},
        "might": ["include"],
        "blocking": False,
        "allow_non_klio_messages": False,
    }


@pytest.fixture
def pipeline_config_dict():
    return {
        "project": "test-project",
        "staging_location": "gs://some/stage",
        "temp_location": "gs://some/temp",
        "worker_harness_container_image": "gcr.io/sigint/foo",
        "streaming": True,
        "update": False,
        "experiments": ["beam_fn_api"],
        "region": "us-central1",
        "num_workers": 3,
        "random_param": "wombat",
        "max_num_workers": 5,
        "disk_size_gb": 50,
        "worker_machine_type": "n1-standard-4",
        "dataflow_endpoint": "https://example.com",
        "service_account_email": "sudo@example.com",
        "no_auth": True,
        "template_location": "gs://some/template",
        "labels": ["some", "labels"],
        "label": "single_label",
        "transform_name_mapping": '{"transform": "is mapped", "using": "json"}',
        "dataflow_kms_key": "a_key_name",
        "autoscaling_algorithm": "THROUGHPUT_BASED",
        "flexrs_goal": "COST_OPTIMIZED",
        "worker_disk_type": "pd-ssd",
        "use_public_ips": True,
        "min_cpu_platform": "Intel Skylake",
        "dataflow_worker_jar": "/foo/bar.jar",
        "subnetwork": "https://www.googleapis.com/compute/v1/projects/"
        "test-project/regions/us-central1/subnetworks/xpn-us1",
    }


@pytest.fixture
def final_pipeline_config_dict():
    return {
        "project": "test-project",
        "staging_location": "gs://some/stage",
        "temp_location": "gs://some/temp",
        "worker_harness_container_image": "gcr.io/sigint/foo",
        "streaming": True,
        "update": False,
        "experiments": ["beam_fn_api"],
        "region": "us-central1",
        "num_workers": 3,
        "random_param": "wombat",
        "max_num_workers": 5,
        "disk_size_gb": 50,
        "worker_machine_type": "n1-standard-4",
        "dataflow_endpoint": "https://example.com",
        "service_account_email": "sudo@example.com",
        "no_auth": True,
        "template_location": "gs://some/template",
        "labels": ["some", "labels", "single_label"],
        "label": "single_label",
        "transform_name_mapping": '{"transform": "is mapped", "using": "json"}',
        "dataflow_kms_key": "a_key_name",
        "flexrs_goal": "COST_OPTIMIZED",
        "worker_disk_type": "compute.googleapis.com/projects/"
        "test-project/regions/us-central1/diskTypes/pd-ssd",
        "use_public_ips": True,
        "min_cpu_platform": "Intel Skylake",
        "dataflow_worker_jar": "/foo/bar.jar",
        "job_name": "test-job",
        "runner": "DataflowRunner",
        "subnetwork": "https://www.googleapis.com/compute/v1/projects/"
        "test-project/regions/us-central1/subnetworks/xpn-us1",
        "enable_streaming_engine": False,
        "autoscaling_algorithm": "THROUGHPUT_BASED",
        "requirements_file": None,
        "sdk_location": None,
        "setup_file": None,
        "profile_location": None,
        "profile_cpu": None,
        "profile_memory": None,
        "profile_sample_rate": None,
    }


@pytest.fixture
def empty_pipeline_config_dict():
    return {
        "project": "test-project",
        "staging_location": "gs://some/stage",
        "temp_location": "gs://some/temp",
        "worker_harness_container_image": "gcr.io/sigint/foo",
        "streaming": True,
        "update": False,
        "experiments": ["beam_fn_api"],
        "region": "us-central1",
        "num_workers": 3,
        "max_num_workers": 5,
        "disk_size_gb": 50,
        "worker_machine_type": "n1-standard-4",
    }


@pytest.fixture
def bare_pipeline_config_dict():
    return {
        "streaming": True,
        "experiments": ["beam_fn_api"],
        "runner": "direct",
    }


@pytest.fixture
def config_dict(job_config_dict, pipeline_config_dict):
    return {
        "job_config": job_config_dict,
        "pipeline_options": pipeline_config_dict,
        "job_name": "test-job",
    }


@pytest.fixture
def final_config_dict(final_pipeline_config_dict, final_job_config_dict):
    return {
        "version": 2,
        "job_name": "test-job",
        "pipeline_options": final_pipeline_config_dict,
        "job_config": final_job_config_dict,
    }


@pytest.fixture
def no_gcp_config_dict(job_config_dict, empty_pipeline_config_dict):
    return {
        "job_config": job_config_dict,
        "pipeline_options": empty_pipeline_config_dict,
        "job_name": "test-job",
    }


@pytest.mark.parametrize("blocking", (True, False, None))
def test_klio_job_config(
    job_config_dict, blocking, final_job_config_dict,
):
    if blocking is None:
        job_config_dict.pop("blocking")
    else:
        job_config_dict["blocking"] = blocking
        final_job_config_dict["blocking"] = blocking

    config_obj = config.KlioJobConfig(
        job_config_dict, job_name="test-job", version=2
    )

    assert {"logger": {}} == config_obj.metrics

    ret_input_topics = [i.topic for i in config_obj.events.inputs]
    ret_output_topics = [o.topic for o in config_obj.events.outputs]

    assert ["test-parent-job-out"] == ret_input_topics
    assert ["test-job-out"] == ret_output_topics

    ret_input_data = [i.location for i in config_obj.data.inputs]
    ret_output_data = [o.location for o in config_obj.data.outputs]

    assert ["gs://sigint-output/test-parent-job-out"] == ret_input_data
    assert ["gs://sigint-output/test-job-out"] == ret_output_data

    if blocking:
        assert config_obj.blocking is True
    else:
        assert config_obj.blocking is False

    if blocking is None:
        job_config_dict["blocking"] = False

    assert final_job_config_dict == config_obj.as_dict()

    repr_actual = repr(config_obj)
    assert "KlioJobConfig(job_name='test-job')" == repr_actual


def test_bare_klio_pipeline_config(bare_pipeline_config_dict):
    config_obj = config.KlioPipelineConfig(
        bare_pipeline_config_dict, version=1, job_name="test-job"
    )

    assert config_obj.streaming is True
    assert not config_obj.update
    assert ["beam_fn_api"] == config_obj.experiments
    assert "direct" == config_obj.runner
    assert "europe-west1" == config_obj.region
    assert config_obj.subnetwork is None
    assert 2 == config_obj.num_workers
    assert 2 == config_obj.max_num_workers
    assert 32 == config_obj.disk_size_gb
    assert "n1-standard-2" == config_obj.worker_machine_type
    assert config_obj.no_auth is False
    assert [] == config_obj.labels
    assert config_obj.enable_streaming_engine is False
    assert config_obj.autoscaling_algorithm == "NONE"

    expected_none_attrs = [
        "project",
        "staging_location",
        "temp_location",
        "worker_harness_container_image",
        "sdk_location",
        "setup_file",
        "requirements_file",
        "dataflow_endpoint",
        "service_account_email",
        "template_location",
        "transform_name_mapping",
        "dataflow_kms_key",
        "flexrs_goal",
        "worker_disk_type",
        "use_public_ips",
        "min_cpu_platform",
        "dataflow_worker_jar",
    ]
    for attr in expected_none_attrs:
        attr_to_test = getattr(config_obj, attr)
        assert attr_to_test is None

    repr_actual = repr(config_obj)
    assert "KlioPipelineConfig(job_name='test-job')" == repr_actual


def test_klio_pipeline_config(
    pipeline_config_dict, final_pipeline_config_dict,
):

    config_obj = config.KlioPipelineConfig(
        pipeline_config_dict, job_name="test-job", version=1
    )

    config_sub_network = (
        "https://www.googleapis.com/compute/v1/projects/test-project/"
        "regions/us-central1/subnetworks/xpn-us1"
    )

    assert "test-project" == config_obj.project
    assert "gs://some/stage" == config_obj.staging_location
    assert "gs://some/temp" == config_obj.temp_location
    assert "gcr.io/sigint/foo" == config_obj.worker_harness_container_image
    assert config_obj.streaming
    assert not config_obj.update
    assert ["beam_fn_api"] == config_obj.experiments
    assert "us-central1" == config_obj.region
    assert config_sub_network == config_obj.subnetwork
    assert 3 == config_obj.num_workers
    assert 5 == config_obj.max_num_workers
    assert 50 == config_obj.disk_size_gb
    assert "n1-standard-4" == config_obj.worker_machine_type
    assert config_obj.sdk_location is None
    assert "DataflowRunner" == config_obj.runner
    assert "https://example.com" == config_obj.dataflow_endpoint
    assert "sudo@example.com" == config_obj.service_account_email
    assert config_obj.no_auth is True
    assert "gs://some/template" == config_obj.template_location
    assert ["some", "labels", "single_label"] == config_obj.labels
    assert "single_label" == config_obj.label
    assert (
        '{"transform": "is mapped", "using": "json"}'
        == config_obj.transform_name_mapping
    )
    assert "THROUGHPUT_BASED" == config_obj.autoscaling_algorithm
    assert "COST_OPTIMIZED" == config_obj.flexrs_goal
    assert (
        "compute.googleapis.com/projects/test-project/regions/us-central1/"
        "diskTypes/pd-ssd" == config_obj.worker_disk_type
    )
    assert config_obj.use_public_ips is True
    assert "Intel Skylake" == config_obj.min_cpu_platform
    assert "/foo/bar.jar" == config_obj.dataflow_worker_jar
    assert "wombat" == config_obj.random_param

    assert final_pipeline_config_dict == config_obj.as_dict()

    repr_actual = repr(config_obj)
    assert "KlioPipelineConfig(job_name='test-job')" == repr_actual


def test_klio_config(config_dict, final_config_dict):

    config_obj = config.KlioConfig(config_dict, config_skip_preprocessing=True)

    assert "test-job" == config_obj.job_name
    assert isinstance(config_obj.job_config, config.KlioJobConfig)
    assert isinstance(config_obj.pipeline_options, config.KlioPipelineConfig)
    assert final_config_dict == config_obj.as_dict()

    repr_actual = repr(config_obj)
    assert "KlioConfig(job_name='test-job')" == repr_actual


def test_no_gcp_klio_config(no_gcp_config_dict):

    config_obj = config.KlioConfig(
        no_gcp_config_dict, config_skip_preprocessing=True
    )

    assert "test-job" == config_obj.job_name
    assert isinstance(config_obj.job_config, config.KlioJobConfig)
    assert isinstance(config_obj.pipeline_options, config.KlioPipelineConfig)
    # Default variables are added to the pipeline config
    assert config_dict != config_obj.as_dict()

    repr_actual = repr(config_obj)
    assert "KlioConfig(job_name='test-job')" == repr_actual


def test_klio_read_file_config():
    config_dict = {
        "type": "GCS",
        "location": "gs://sigint-output/test-parent-job-out",
    }
    klio_read_file_config = io.KlioReadFileConfig.from_dict(
        config_dict, io.KlioIOType.DATA, io.KlioIODirection.INPUT
    )

    assert "file" == klio_read_file_config.name
    assert config_dict["location"] == klio_read_file_config.file_pattern


def test_klio_write_file_config():
    config_dict = {
        "type": "GCS",
        "location": "gs://sigint-output/test-parent-job-out",
    }
    klio_write_file_config = io.KlioWriteFileConfig.from_dict(
        config_dict, io.KlioIOType.DATA, io.KlioIODirection.OUTPUT
    )

    assert "file" == klio_write_file_config.name
    assert config_dict["location"] == klio_write_file_config.file_path_prefix


def test_klio_write_bigquery_config():
    config_dict = {
        "type": "bq",
        "project": "a-project",
        "dataset": "a-dataset",
        "table": "a-table",
        "create_disposition": "CREATE_IF_NEEDED",
        "write_disposition": "WRITE_TRUNCATE",
        "schema": {
            "fields": [{"name": "label", "type": "STRING", "mode": "nullable"}]
        },
    }

    klio_write_bq_cfg = io.KlioBigQueryEventOutput.from_dict(
        config_dict, io.KlioIOType.EVENT, io.KlioIODirection.OUTPUT
    )

    assert "bq" == klio_write_bq_cfg.name
    assert config_dict["schema"] == klio_write_bq_cfg.schema
    assert (
        config_dict["create_disposition"]
        == klio_write_bq_cfg.create_disposition
    )
    assert (
        config_dict["write_disposition"] == klio_write_bq_cfg.write_disposition
    )


@pytest.mark.parametrize(
    "schema",
    (
        {},
        {"not-fields": []},
        {"fields": None},
        {"fields": [{}]},
        {"fields": [{"name": "a", "type": "b"}]},
        {"fields": [{"name": "a", "type": "b", "mode": "c"}, {}]},
    ),
)
def test_klio_write_bigquery_config_raises(schema):
    config_dict = {
        "type": "bq",
        "project": "a-project",
        "dataset": "a-dataset",
        "table": "a-table",
        "create_disposition": "CREATE_IF_NEEDED",
        "write_disposition": "WRITE_TRUNCATE",
        "schema": schema,
    }

    with pytest.raises(ValueError):
        io.KlioBigQueryEventOutput.from_dict(
            config_dict, io.KlioIOType.EVENT, io.KlioIODirection.OUTPUT
        )


@pytest.mark.parametrize(
    "include_topic,include_subscription",
    ((False, True), (True, False), (True, True),),
)
def test_pubsub_event_input_kwargs(include_topic, include_subscription):
    config_dict = {
        "type": "pubsub",
        "topic": "a-topic",
        "subscription": "a-subscription",
    }

    if not include_topic:
        config_dict.pop("topic")
    if not include_subscription:
        config_dict.pop("subscription")

    pubsub = io.KlioPubSubEventInput.from_dict(
        config_dict, io.KlioIOType.EVENT, io.KlioIODirection.INPUT
    )

    if include_subscription:
        expected = {
            "subscription": "a-subscription",
        }
    else:
        expected = {
            "topic": "a-topic",
        }

    assert expected == pubsub.to_io_kwargs()


def test_pubsub_event_input_topic_subscription():
    config_dict = {"type": "pubsub"}

    with pytest.raises(ValueError):
        io.KlioPubSubEventInput.from_dict(
            config_dict, io.KlioIOType.EVENT, io.KlioIODirection.INPUT
        )


def test_config_pickling(config_dict, final_config_dict):
    # This test attempts to verify that class-level attributes aren't used as
    # instance attributes, since they are not pickled, which can result in
    # missing/wrong values when config is unpickled on dataflow workers

    def get_class_attributes(cls):
        attrs = {}
        for key in cls.__dict__:
            value = getattr(cls, key)
            is_fn = inspect.ismethod(value) or inspect.isfunction(value)
            if not key.startswith("__") and not is_fn:
                attrs[key] = copy.copy(value)
        return attrs

    classes = [
        config.KlioConfig,
        config.KlioJobConfig,
        config.KlioPipelineConfig,
    ]

    cls_attribs = {}

    for cls in classes:
        cls_attribs[cls] = get_class_attributes(cls)

    klio_config = config.KlioConfig(
        config_dict, config_skip_preprocessing=True
    )

    pickled = dill.dumps(klio_config)

    # reset any class-level attributes back to whatever value they had before
    # instantiating KlioConfig
    for cls, keyvals in cls_attribs.items():
        for key, value in keyvals.items():
            setattr(cls, key, value)

    unpickled = dill.loads(pickled)

    actual = unpickled.as_dict()

    assert final_config_dict == actual


@pytest.mark.parametrize(
    "worker_disk_type, is_valid",
    (
        ("local-ssd", True),
        ("something", False),
        (
            "compute.googleapis.com/projects/test_project/"
            "regions/europe-west1/diskTypes/local-ssd",
            True,
        ),
        (
            "compute.googleapis.com/projects/wrong_project/"
            "regions/europe-west1/diskTypes/local-ssd",
            False,
        ),
        (
            "compute.googleapis.com/projects/test_project/"
            "regions/wrong-region/diskTypes/local-ssd",
            False,
        ),
        (
            "compute.googleapis.com/projects/test_project/"
            "regions/wrong-region/diskTypes/invalid",
            False,
        ),
    ),
)
def test_worker_disk_image_formatting(worker_disk_type, is_valid):
    # config formats worker_disk_type on creation.  We need to ensure it can
    # read the unformatted and formatted versions
    pipeline_config_dict = {
        "project": "test_project",
        "worker_disk_type": worker_disk_type,
    }

    if is_valid:
        config.KlioPipelineConfig(
            pipeline_config_dict, job_name="test_job", version=2
        )
    else:
        with pytest.raises(ValueError):
            config.KlioPipelineConfig(
                pipeline_config_dict, job_name="test_job", version=2
            )



# Copyright 2019-2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import enum
import importlib
import json
import logging


import attr

logger = logging.getLogger("klio")


class KlioIODirection(enum.Enum):
    INPUT = 1
    OUTPUT = 2


class KlioIOType(enum.Enum):
    EVENT = 1
    DATA = 2


class KlioJobMode(enum.Enum):
    STREAMING = 1
    BATCH = 2


def supports(*type_directions):
    """Decorator for subclasses of KlioIOConfig to indicate what IO types and
    directions the Config supports.  In many cases the same config object can
    be used for configuring either input or output, and likewise event I/O and
    data I/O.  Adding respective `KlioIOType` and `KlioIODirection` values via
    this decorator will inform Klio about which class to use when parsing
    config.

    """

    def wrapper(cls):
        io_types = [t for t in type_directions if isinstance(t, KlioIOType)]
        io_directions = [
            t for t in type_directions if isinstance(t, KlioIODirection)
        ]
        # IMPORTANT - new lists must be created here, otherwise all subclasses
        # end up sharing the same lists
        cls.SUPPORTED_TYPES = [t for t in cls.SUPPORTED_TYPES]
        cls.SUPPORTED_DIRECTIONS = [d for d in cls.SUPPORTED_DIRECTIONS]
        for t in io_types:
            if t not in cls.SUPPORTED_TYPES:
                cls.SUPPORTED_TYPES.append(t)
        for d in io_directions:
            if d not in cls.SUPPORTED_DIRECTIONS:
                cls.SUPPORTED_DIRECTIONS.append(d)
        return cls

    return wrapper


@attr.attrs
class KlioIOConfig(object):
    io_type = attr.attrib(type=KlioIOType)
    io_direction = attr.attrib(type=KlioIODirection)

    # these must be filled in by subclasses so Klio knows what supports what
    SUPPORTED_TYPES = []
    SUPPORTED_DIRECTIONS = []
    # NOTICE! any new `attr.attrib`s added to this class should be added in
    # `ATTRIBS_TO_SKIP` below in order to be filtered out when calling
    # `as_dict`, leaving only `attr.attrib`s related to the specific
    # config instance
    ATTRIBS_TO_SKIP = ["io_type", "io_direction"]

    @classmethod
    def supports_type(cls, io_type):
        return io_type in cls.SUPPORTED_TYPES

    @classmethod
    def supports_direction(cls, io_direction):
        return io_direction in cls.SUPPORTED_DIRECTIONS

    @classmethod
    def from_dict(cls, config_dict, *args, **kwargs):
        copy = config_dict.copy()
        if "type" in copy:
            del copy["type"]
        return cls(*args, **copy, **kwargs)

    def _as_dict(self):
        """Return a dictionary-representation of the config object.

        Useful for instantiating objects where the config keys map 1:1
        to the object init arguments/keyword arguments, i.e.
        KlioReadFromBigQuery(**config.job_config.events.inputs[0].as_dict())
        """
        # filter parameter is used as "to include"/"filter-in",
        # not "to exclude"/"filter-out"
        return attr.asdict(
            self, filter=lambda x, _: x.name not in self.ATTRIBS_TO_SKIP
        )

    def as_dict(self):
        config_dict = self._as_dict()
        # since dicts preserve order by default in py3, let's force
        # type to be first - particularly helpful/useful for dumping
        # config via `klio job config show`
        copy = {"type": self.name}
        copy.update(config_dict)
        return copy

    def to_io_kwargs(self):
        return self._as_dict()


@attr.attrs(frozen=True)
class IOFlags(object):
    io_direction = attr.attrib(type=KlioIODirection)
    io_type = attr.attrib(type=KlioIOType)
    job_mode = attr.attrib(type=KlioJobMode)


@attr.attrs(frozen=True)
class KlioEventInput(KlioIOConfig):
    skip_klio_read = attr.attrib(type=bool)

    @classmethod
    def from_dict(cls, config_dict, *args, **kwargs):
        # work-around with attrs - since attrs will not allow attributes
        # to be defined without a default after those that have defaults,
        # we're inserting the default value here if the user doesn't have
        # it already in their config
        if "skip_klio_read" not in config_dict:
            copy = config_dict.copy()
            copy["skip_klio_read"] = False
            return super().from_dict(copy, *args, **kwargs)
        return super().from_dict(config_dict, *args, **kwargs)

    def to_io_kwargs(self):
        kwargs = super().to_io_kwargs()
        kwargs.pop("skip_klio_read", None)
        return kwargs


@attr.attrs(frozen=True)
class KlioEventOutput(KlioIOConfig):
    skip_klio_write = attr.attrib(type=bool)

    @classmethod
    def from_dict(cls, config_dict, *args, **kwargs):
        # work-around with attrs - since attrs will not allow attributes
        # to be defined without a default after those that have defaults,
        # we're inserting the default value here if the user doesn't have
        # it already in their config
        if "skip_klio_write" not in config_dict:
            copy = config_dict.copy()
            copy["skip_klio_write"] = False
            return super().from_dict(copy, *args, **kwargs)
        return super().from_dict(config_dict, *args, **kwargs)

    def to_io_kwargs(self):
        kwargs = super().to_io_kwargs()
        kwargs.pop("skip_klio_write", None)
        return kwargs


@attr.attrs(frozen=True)
class KlioDataIOConfig(KlioIOConfig):
    skip_klio_existence_check = attr.attrib(type=bool)

    @classmethod
    def from_dict(cls, config_dict, *args, **kwargs):
        # work-around with attrs - since attrs will not allow attributes
        # to be defined without a default after those that have defaults,
        # we're inserting the default value here if the user doesn't have
        # it already in their config
        if "skip_klio_existence_check" not in config_dict:
            copy = config_dict.copy()
            copy["skip_klio_existence_check"] = False
            return super().from_dict(copy, *args, **kwargs)
        return super().from_dict(config_dict, *args, **kwargs)

    def to_io_kwargs(self):
        kwargs = super().to_io_kwargs()
        kwargs.pop("skip_klio_existence_check", None)
        return kwargs


@attr.attrs(frozen=True)
class KlioPubSubConfig(object):
    name = "pubsub"
    topic = attr.attrib(type=str)

    @staticmethod
    def _from_dict(config_dict):
        copy = config_dict.copy()
        copy.pop("data_location", None)
        return copy


@supports(KlioIODirection.INPUT, KlioIOType.EVENT)
@attr.attrs(frozen=True)
class KlioPubSubEventInput(KlioEventInput, KlioPubSubConfig):
    subscription = attr.attrib(type=str)

    @classmethod
    def from_dict(cls, config_dict, *args, **kwargs):
        config_dict = super()._from_dict(config_dict)
        # work-around with attrs - since attrs will not allow attributes
        # to be defined without a default after those that have defaults,
        # we're inserting the default value here if the user doesn't have
        # it already in their config
        if "topic" not in config_dict:
            config_dict["topic"] = None
        if "subscription" not in config_dict:
            config_dict["subscription"] = None
        return super().from_dict(config_dict, *args, **kwargs)

    @subscription.validator
    def __assert_topic_subscription(self, attribute, value):
        # either topic or subscription is required
        if not self.topic and not self.subscription:
            raise ValueError("One of 'topic', 'subscription' required")

    def to_io_kwargs(self):
        kwargs = super().to_io_kwargs()
        # pubsub only allows either topic or subscription, not both.  We'll
        # prioritize subscription
        if kwargs["topic"] is not None and kwargs["subscription"] is None:
            kwargs.pop("subscription", None)
        else:
            kwargs.pop("topic", None)
        return kwargs


@supports(KlioIODirection.OUTPUT, KlioIOType.EVENT)
@attr.attrs(frozen=True)
class KlioPubSubEventOutput(KlioEventOutput, KlioPubSubConfig):
    @classmethod
    def from_dict(cls, config_dict, *args, **kwargs):
        config_dict = super()._from_dict(config_dict)
        return super().from_dict(config_dict, *args, **kwargs)


class KlioFileConfig(object):
    name = "file"


@attr.attrs(frozen=True)
class KlioReadFileConfig(KlioEventInput, KlioFileConfig):
    file_pattern = attr.attrib(type=str)

    @classmethod
    def from_dict(cls, config_dict, *args, **kwargs):
        if "location" in config_dict:
            copy = config_dict.copy()
            copy["file_pattern"] = copy.pop("location", None)
            return super().from_dict(copy, *args, **kwargs)
        try:
            return super().from_dict(config_dict, *args, **kwargs)
        except TypeError as e:
            if "file_pattern" in str(e):
                raise KeyError(
                    "I/O configuration for `type: file` is missing 'location' "
                    "key."
                )

    def as_dict(self):
        config_dict = super().as_dict()
        copy = config_dict.copy()
        copy["location"] = copy.pop("file_pattern", None)
        return copy


@attr.attrs(frozen=True)
class KlioWriteFileConfig(KlioEventOutput, KlioFileConfig):
    file_path_prefix = attr.attrib(type=str)

    @classmethod
    def from_dict(cls, config_dict, *args, **kwargs):
        if "location" in config_dict:
            copy = config_dict.copy()
            copy["file_path_prefix"] = copy.pop("location", None)
            return super().from_dict(copy, *args, **kwargs)
        return super().from_dict(config_dict, *args, **kwargs)

    def as_dict(self):
        config_dict = super().as_dict()
        copy = config_dict.copy()
        copy["location"] = copy.pop("file_path_prefix", None)
        return copy


@supports(KlioIODirection.INPUT, KlioIOType.EVENT)
@attr.attrs(frozen=True)
class KlioReadFileEventConfig(KlioReadFileConfig):
    pass


@supports(KlioIODirection.OUTPUT, KlioIOType.EVENT)
@attr.attrs(frozen=True)
class KlioWriteFileEventConfig(KlioWriteFileConfig):
    pass


@attr.attrs(frozen=True)
@supports(KlioIODirection.INPUT, KlioIOType.DATA)
class KlioFileInputDataConfig(KlioDataIOConfig, KlioFileConfig):
    location = attr.attrib(type=str)
    ping = attr.attrib(type=bool, default=False)
    file_suffix = attr.attrib(type=str, default="")


@attr.attrs(frozen=True)
@supports(KlioIODirection.OUTPUT, KlioIOType.DATA)
class KlioFileOutputDataConfig(KlioDataIOConfig, KlioFileConfig):
    location = attr.attrib(type=str)
    file_suffix = attr.attrib(type=str, default="")
    force = attr.attrib(type=bool, default=False)


class KlioAvroConfig(object):
    name = "avro"


@attr.attrs(frozen=True)
class KlioReadAvroConfig(KlioAvroConfig):
    # TODO: Add validation
    # Potentially could set a general file pattern default, i.e. *.avro, too
    # but either file_pattern or location must be set
    file_pattern = attr.attrib(type=str, default=None)
    location = attr.attrib(type=str, default=None)
    min_bundle_size = attr.attrib(type=int, default=0)
    validate = attr.attrib(type=bool, default=True)


@supports(KlioIODirection.INPUT, KlioIOType.EVENT)
@attr.attrs(frozen=True)
class KlioReadAvroEventConfig(KlioEventInput, KlioReadAvroConfig):
    pass


@attr.attrs(frozen=True)
class KlioWriteAvroConfig(KlioAvroConfig):
    # Either file_path_prefix or location must be set
    # Right now `schema` is also allowed to pass inbut anything other than
    # the default schema cannot make proper use of `write_record`
    file_path_prefix = attr.attrib(type=str, default=None)
    location = attr.attrib(type=str, default=None)
    codec = attr.attrib(type=str, default="deflate")
    file_name_suffix = attr.attrib(type=str, default="")
    num_shards = attr.attrib(type=int, default=0)
    shard_name_template = attr.attrib(type=str, default=None)
    mime_type = attr.attrib(type=str, default="application/x-avro")


@supports(KlioIODirection.OUTPUT, KlioIOType.EVENT)
@attr.attrs(frozen=True)
class KlioWriteAvroEventConfig(KlioEventOutput, KlioWriteAvroConfig):
    pass


# TODO: integrate into @dsimon's converter logic once his PR#154 is merged
def _convert_bigquery_input_coder(coder_str):
    # direct runner seems to call this multiple times, prob with pickling;
    # subsequent times seem to call with a None value (not sure why...)
    if coder_str is None:
        return

    coder_path_stems = coder_str.split(".")
    coder_kls_str = coder_path_stems.pop()
    module_str = ".".join(coder_path_stems)
    module = importlib.import_module(module_str)  # should raise if not avail
    return getattr(module, coder_kls_str)  # should raise if not avail


@attr.attrs(frozen=True)
class KlioBigQueryConfig(object):
    name = "bq"
    project = attr.attrib(type=str, default=None)
    dataset = attr.attrib(type=str, default=None)
    table = attr.attrib(type=str, default=None)


@attr.attrs(frozen=True)
@supports(KlioIODirection.INPUT, KlioIOType.EVENT)
class KlioBigQueryEventInput(KlioEventInput, KlioBigQueryConfig):
    # Set to true by default to fail early (check table existence before
    # starting pipeline) (Beam sets it to false by default but suggests it
    # to be set to true in most scenarios)
    validate = attr.attrib(type=bool, default=True)
    # TODO: use @dsimon's converter logic once his PR#154 is merged in
    coder = attr.attrib(
        type=str, default=None, converter=_convert_bigquery_input_coder
    )
    kms_key = attr.attrib(type=str, default=None)
    # Optional, but only applies to project+dataset+table;
    # Klio handles filtering data after a `SELECT *` query
    klio_message_columns = attr.attrib(type=list, default=None)
    # Mutually exclusive with project+dataset+table & klio_message_columns
    query = attr.attrib(type=str, default=None)
    use_standard_sql = attr.attrib(type=bool, default=False)
    flatten_results = attr.attrib(type=bool, default=True)

    @klio_message_columns.validator
    def __assert_project_dataset_table(self, attribute, value):
        has_project_dataset_table = all(
            [self.project, self.dataset, self.table]
        )
        if value is not None:
            if not has_project_dataset_table:
                raise ValueError(
                    "Must include `project`, `dataset` and `table` if "
                    "selecting `columns` for BigQuery."
                )

    @query.validator
    def __assert_query_or_project_dataset_table(self, attribute, value):
        has_project_dataset_table = any(
            [self.project, self.dataset, self.table]
        )
        if value is not None and has_project_dataset_table:
            raise ValueError(
                "`query` is mutually exclusive with `project`, `dataset` "
                "and `table`."
            )

    @classmethod
    def from_dict(cls, config_dict, *args, **kwargs):
        if "columns" in config_dict:
            copy = config_dict.copy()
            copy["klio_message_columns"] = copy.pop("columns", None)
            return super().from_dict(copy, *args, **kwargs)
        return super().from_dict(config_dict, *args, **kwargs)

    def as_dict(self):
        config_dict = super().as_dict()
        copy = config_dict.copy()
        copy["columns"] = copy.pop("klio_message_columns", None)
        return copy


def _convert_bigquery_output_schema(schema):
    if isinstance(schema, dict):
        return schema
    return json.loads(schema)


@attr.attrs(frozen=True)
@supports(KlioIODirection.OUTPUT, KlioIOType.EVENT)
class KlioBigQueryEventOutput(KlioEventOutput, KlioBigQueryConfig):
    # schema field is optional; assumes form of
    # {"fields": [{"name": ...,"type": ..., "mode": ...}, ... ]
    schema = attr.attrib(
        type=dict, converter=_convert_bigquery_output_schema, default=None
    )
    create_disposition = attr.attrib(type=str, default="CREATE_IF_NEEDED")
    write_disposition = attr.attrib(type=str, default="WRITE_EMPTY")

    @schema.validator
    def check(self, attribute, value):
        def valid_field(field_dict):
            is_dict = isinstance(field_dict, dict)
            required_keys = [
                "name",
                "type",
                "mode",
            ]  # are all of these required?
            return is_dict and all(k in field_dict for k in required_keys)

        def contains_invalid_fields(field_list):
            return any(valid_field(f) is False for f in field_list)

        has_fields = value.get("fields")
        if not has_fields or contains_invalid_fields(value.get("fields", [])):
            raise ValueError(
                "Must be a dict with the key `fields` set to a list of"
                "dict, each of which have the keys "
                "`name`, `type`, and `mode`"
            )


@attr.attrs(frozen=True)
class KlioGCSConfig(KlioIOConfig):
    name = "gcs"
    location = attr.attrib(type=str)

    @staticmethod
    def _from_dict(config_dict):
        copy = config_dict.copy()
        copy.pop("topic", None)
        copy.pop("subscription", None)
        data_location = copy.pop("data_location", None)
        copy["location"] = copy.get("location", data_location)
        return copy


@attr.attrs(frozen=True)
@supports(KlioIODirection.INPUT, KlioIOType.DATA)
class KlioGCSInputDataConfig(KlioDataIOConfig, KlioGCSConfig):
    file_suffix = attr.attrib(type=str, default="")
    ping = attr.attrib(type=bool, default=False)

    @classmethod
    def from_dict(cls, config_dict, *args, **kwargs):
        config_dict = super()._from_dict(config_dict)
        return super().from_dict(config_dict, *args, **kwargs)


@attr.attrs(frozen=True)
@supports(KlioIODirection.OUTPUT, KlioIOType.DATA)
class KlioGCSOutputDataConfig(KlioDataIOConfig, KlioGCSConfig):
    file_suffix = attr.attrib(type=str, default="")
    force = attr.attrib(type=bool, default=False)

    @classmethod
    def from_dict(cls, config_dict, *args, **kwargs):
        config_dict = super()._from_dict(config_dict)
        return super().from_dict(config_dict, *args, **kwargs)



# Copyright 2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import click


class MutuallyExclusiveOption(click.Option):
    """
    Helper class to validate and document mutual exclusivity between options.

    This unfortunately doesn't work with click arguments (only options).

    To use, pass in both of the following keywords into an option declaration:
        click.option(
            "--an-option"
            cls=MutuallyExclusiveOption
            mutually_exclusive=["string_of_exclusive_option", "another_option"]
        )
    """

    def __init__(self, *args, **kwargs):
        self.mut_ex_opts = set(kwargs.pop("mutually_exclusive", []))
        help_text = kwargs.get("help", "")
        if self.mut_ex_opts:
            mutex = [self._varname_to_opt_flag(m) for m in self.mut_ex_opts]
            mutex_fmted = ["``{}``".format(m) for m in mutex]
            ex_str = ", ".join(mutex_fmted)
            kwargs["help"] = help_text + (
                "\n\n**NOTE:** This option is mutually exclusive with "
                "[{}].".format(ex_str)
            )
        super(MutuallyExclusiveOption, self).__init__(*args, **kwargs)

    @staticmethod
    def _varname_to_opt_flag(var):
        return "--" + var.replace("_", "-")

    def handle_parse_result(self, ctx, opts, args):
        if self.mut_ex_opts.intersection(opts) and self.name in opts:
            mutex = [
                "`" + self._varname_to_opt_flag(m) + "`"
                for m in self.mut_ex_opts
            ]
            mutex = ", ".join(mutex)
            msg = "Illegal usage: `{}` is mutually exclusive with {}.".format(
                self._varname_to_opt_flag(self.name), mutex
            )
            raise click.UsageError(msg)

        return super(MutuallyExclusiveOption, self).handle_parse_result(
            ctx, opts, args
        )


def image_tag(*args, **kwargs):
    """Docker image tag to use."""

    def wrapper(func):
        return click.option(
            "--image-tag", help="Docker image tag to use", **kwargs
        )(func)

    if args:
        return wrapper(args[0])
    return wrapper


#####
# common options
#####
def direct_runner(func):
    """Run the job locally via the DirectRunner."""
    return click.option(
        "--direct-runner",
        default=False,
        is_flag=True,
        help="Run the job locally via the DirectRunner.",
    )(func)


def update(func):
    """[Experimental] Update an existing streaming Cloud Dataflow job."""
    return click.option(
        "--update/--no-update",
        default=None,
        is_flag=True,
        help="[Experimental] Update an existing streaming Cloud Dataflow job.",
    )(func)


def show_logs(func):
    """Show a job's logs while profiling."""
    return click.option(
        "--show-logs",
        default=False,
        show_default=True,
        is_flag=True,
        help="Show a job's logs while profiling.",
    )(func)


#####
# options for `<cmd> profile memory`
#####
def interval(func):
    """Sampling period (in seconds)."""
    return click.option(
        "--interval",
        default=0.1,
        show_default=True,
        type=float,
        help="Sampling period (in seconds).",
    )(func)


def include_children(func):
    """Monitor forked processes as well (sums up all process memory)."""
    return click.option(
        "--include-children",
        default=False,
        show_default=True,
        is_flag=True,
        help="Monitor forked processes as well (sums up all process memory).",
    )(func)


def multiprocess(func):
    """Monitor forked processes creating individual plots for each child."""
    return click.option(
        "--multiprocess",
        default=False,
        show_default=True,
        is_flag=True,
        help=(
            "Monitor forked processes creating individual plots for each "
            "child."
        ),
    )(func)


def plot_graph(func):
    """
    Plot memory profile using matplotlib.
    Saves to klio_profile_memory_<YYYYMMDDhhmmss>.png
    """
    return click.option(
        "-g",
        "--plot-graph",
        default=False,
        show_default=True,
        is_flag=True,
        help=(
            "Plot memory profile using matplotlib. Saves to "
            "klio_profile_memory_<YYYYMMDDhhmmss>.png."
        ),
    )(func)


#####
# options for `<cmd> profile memory-per-line`
#####
def maximum(func):
    """
    Print maximum memory usage per line in aggregate of all input
    elements processed
    """
    return click.option(
        "--maximum",
        "get_maximum",
        default=False,
        show_default=True,
        is_flag=True,
        cls=MutuallyExclusiveOption,
        mutually_exclusive=["per_element"],
        help=(
            "Print maximum memory usage per line in aggregate of all input "
            "elements process."
        ),
    )(func)


def per_element(*args, **kwargs):
    """Print memory useage per line for each input element processed"""

    def wrapper(func):
        return click.option(
            "--per-element",
            default=False,
            is_flag=True,
            cls=MutuallyExclusiveOption,
            mutually_exclusive=["maximum"],
            help="Print memory usage per line for each input element processed",
            **kwargs  # added (and no more show_default keyword arg)
        )(func)

    if args:
        return wrapper(args[0])
    return wrapper


#####
# options for `<cmd> profile timeit`
#####
def iterations(func):
    """Number of times to execute each entity ID provided."""
    return click.option(
        "-n",
        "--iterations",
        default=10,
        show_default=True,
        type=int,
        help="Number of times to execute each entity ID provided.",
    )(func)


################################
# options for config overrides
################################


def override(func):
    """Override a config value, in the form ``key=value``."""
    return click.option(
        "-O",
        "--override",
        default=[],
        multiple=True,
        help="Override a config value, in the form ``key=value``.",
    )(func)


def template(func):
    """
    Set the value of a config template parameter in the form ``key=value``.
    Any instance of ``${key}`` in ``klio-job.yaml`` will be replaced
    with ``value``.
    """
    return click.option(
        "-T",
        "--template",
        default=[],
        multiple=True,
        help=(
            "Set the value of a config template parameter"
            ", in the form ``key=value``.  Any instance of ``${key}`` "
            "in ``klio-job.yaml`` will be replaced with ``value``."
        ),
    )(func)


def config_file(*args, **kwargs):
    """Path to config filename. If ``PATH`` is not absolute,
    it will be treated relative to ``--job-dir``.
    Defaults to ``klio-job.yaml``.
    """
    mutually_exclusive = kwargs.get("mutex", [])

    def wrapper(func):
        return click.option(
            "-c",
            "--config-file",
            type=click.Path(exists=False),
            help=(
                "Path to config filename. If ``PATH`` is not absolute, it "
                "will be treated relative to ``--job-dir``. Defaults to "
                "``klio-job.yaml``."
            ),
            cls=MutuallyExclusiveOption,
            mutually_exclusive=mutually_exclusive,
        )(func)

    # allows @options.foo to be used without parens (i.e. no need to do
    # `@options.foo()`) when there are no args/kwargs provided
    if args:
        return wrapper(args[0])
    return wrapper


def job_dir(*args, **kwargs):
    """
    Job directory where the job's ``Dockerfile`` is located.
    Defaults current working directory.
    """
    mutually_exclusive = kwargs.get("mutex", [])

    def wrapper(func):
        return click.option(
            "-j",
            "--job-dir",
            type=click.Path(exists=True),
            help=(
                "Job directory where the job's ``Dockerfile`` is located. "
                "Defaults current working directory."
            ),
            cls=MutuallyExclusiveOption,
            mutually_exclusive=mutually_exclusive,
        )(func)

    # allows @options.foo to be used without parens (i.e. no need to do
    # `@options.foo()`) when there are no args/kwargs provided
    if args:
        return wrapper(args[0])
    return wrapper



# Copyright 2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Common exceptions for use within the Klio ecosystem."""


class KlioConfigTemplatingException(Exception):
    """Exception for missing keys when overriding Klio config parameters

    Args:
        key (str): missing configuration key
    """

    def __init__(self, key):
        message = "{} missing in key template overrides.".format(key)
        super().__init__(message)



# Copyright 2019-2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Module for interacting with the `Dataflow REST API
<https://cloud.google.com/dataflow/docs/reference/rest>`_."""

import functools
import logging
import random

from googleapiclient import discovery

from klio_core import utils
from klio_core import variables


class DataflowClient(object):
    """Client to interact with Dataflow REST API.

    Args:
        api_version (str): Version of Dataflow REST API. Defaults to
            ``v1b3``.
    """

    def __init__(self, api_version=None):
        _api_version = api_version or "v1b3"
        self.client = discovery.build("dataflow", _api_version)
        self.logger = logging.getLogger("klio")

    def find_job_by_name(self, job_name, gcp_project, region=None):
        """Search Dataflow for a job given its name and GCP project.

        Args:
            job_name (str): Name of Dataflow job.
            gcp_project (str): GCP project in which to search.
            region (str): Region in which to search. Defaults to
                searching all regions in
                :attr:`klio_core.variables.DATAFLOW_REGIONS`.
        Returns:
            dict or None: If found, ``dict`` of job summary results. Otherwise,
            ``None``.
        """
        if not region:
            regions = variables.DATAFLOW_REGIONS
        else:
            regions = (region,)

        base_request = self.client.projects().locations().jobs()

        all_matching_jobs = []

        # TODO: no batch requesting from Google's side, but should add
        #       threading to send multiple requests concurrently. @lynn
        for region in regions:
            # Note: the parameter `view="JOB_VIEW_ALL"` does not return
            #       the same information in this `.list()` call as it
            #       does in the `.get()` call in `get_job_detail` below.
            request = base_request.list(
                projectId=gcp_project, location=region, filter="ACTIVE"
            )

            try:
                response = request.execute()

            # general catch all since the handling would be the same no matter
            # of the exception
            except Exception as e:
                self.logger.warning(
                    "Error listing active jobs in project '%s' in region '%s':"
                    " %s" % (gcp_project, region, e)
                )
                continue

            job_results = response.get("jobs", [])
            if job_results:
                for result in job_results:
                    if result["name"] == job_name:
                        all_matching_jobs.append(result)

        # Note: job names are unique within regions, but not across
        #       regions :grimace:
        if len(all_matching_jobs) > 1:
            self.logger.info(
                "More than one parent job found for job name '%s' under "
                "project '%s'. Selecting one at random."
            )
            return random.choice(all_matching_jobs)
        if all_matching_jobs:
            return all_matching_jobs[0]

    def get_job_detail(self, job_name, gcp_project, region=None):
        """Get verbose job detail given a job name.

        Args:
            job_name (str): Name of Dataflow job.
            gcp_project (str): GCP project in which to search.
            region (str): Region in which to search. Defaults to
                searching all regions in
                :attr:`klio_core.variables.DATAFLOW_REGIONS`.
        Returns:
            dict or None: If found, ``dict`` of detailed job results.
            Otherwise, ``None``.
        """
        basic_job = self.find_job_by_name(job_name, gcp_project, region)
        if not basic_job:
            return None

        job_id = basic_job["id"]
        job_location = basic_job["location"]

        request = (
            self.client.projects()
            .locations()
            .jobs()
            .get(
                projectId=gcp_project,
                location=job_location,
                jobId=job_id,
                view="JOB_VIEW_ALL",
            )
        )
        try:
            response = request.execute()
        # general catch all since the handling would be the same no matter
        # of the exception
        except Exception as e:
            self.logger.warning(
                "Error getting job detail for '%s' in project '%s' in "
                "region '%s': %s" % (job_name, gcp_project, job_location, e)
            )
            return

        return response

    def get_job_input_topic(self, job_name, gcp_project, region=None):
        """Get input topic of a particular job.

        Args:
            job_name (str): Name of Dataflow job.
            gcp_project (str): GCP project in which to search.
            region (str): Region in which to search. Defaults to
                searching all regions in
                :attr:`klio_core.variables.DATAFLOW_REGIONS`.
        Returns:
            str or None: If found, input topic of job. Otherwise, ``None``.
        """
        job_info = self.get_job_detail(job_name, gcp_project, region=None)
        if not job_info:
            return None

        read_pubsub_user_name = "ReadFromPubSub/Read"

        for step in job_info.get("steps", []):
            if step.get("kind") == "ParallelRead":
                props = step.get("properties", {})
                user_name = props.get("user_name", {})
                if user_name.get("value") == read_pubsub_user_name:
                    # TODO: support multiple input topics; will need to
                    #       see how Google's response json renders it. @lynn
                    return props.get("pubsub_topic", {}).get("value")


def get_dataflow_client(api_version=None):
    """Get an initialized :class:`DataflowClient`.

    This function will first check if there is an already initialized
    client in the global namespace. Otherwise, initialize one then set it
    in the global namespace to avoid redundant initialization.

    Args:
        api_version (str): Version of Dataflow REST API. Defaults to
            ``v1b3``.
    Returns:
        DataflowClient: A client to interact with the Dataflow REST API.
    """
    if not api_version:
        api_version = "v1b3"
    key = "dataflow_client_{}".format(api_version)
    initializer = functools.partial(DataflowClient, api_version)
    return utils.get_or_initialize_global(key, initializer)



# Copyright 2019-2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import pytest

from klio_core import config
from klio_core.config import _converters as converters
from klio_core.config import _utils as utils


@utils.config_object(key_prefix="foo.bar")
class ConfigTestClass(object):
    f1 = utils.field(type=str, default=None)
    f2 = utils.field(type=str)


def test_config_decorator_direct_instantiation():
    a = ConfigTestClass(f1="f1value", f2=None)
    assert "f1value" == a.f1
    assert a.f2 is None

    b = ConfigTestClass(f2="value")
    assert b.f1 is None
    assert "value" == b.f2

    with pytest.raises(Exception):
        ConfigTestClass(f1="value")


@pytest.mark.parametrize(
    "config_dict, expected",
    (
        (
            {"f1": "f1value", "f2": None},
            ConfigTestClass(f1="f1value", f2=None),
        ),
        ({"f1": "f1value"}, None),
        ({"f2": "value"}, ConfigTestClass(f1=None, f2="value")),
        ({}, None),
    ),
)
def test_config_decorator_no_value(config_dict, expected):
    if expected:
        assert expected == ConfigTestClass(config_dict)
    else:
        with pytest.raises(Exception):
            ConfigTestClass(config_dict)


@pytest.mark.parametrize(
    "value, expected", ((5, "5"), ("foo", "foo"), (None, None), (True, "True"))
)
def test_str_converter(value, expected):
    assert expected == converters.StringConverter("foo").validate(value)


@pytest.mark.parametrize(
    "bad_value", (converters.UNSET_REQUIRED_VALUE, {}, [])
)
def test_str_converter_raises(bad_value):
    with pytest.raises(Exception):
        converters.StringConverter("foo").validate(bad_value)


@pytest.mark.parametrize(
    "value, expected",
    (
        (5, True),
        (True, True),
        (None, None),
        (0, False),
        (False, False),
        ("true", True),
        ("false", True),  # hmmmmm
    ),
)
def test_bool_converter(value, expected):
    assert expected == converters.BoolConverter("foo").validate(value)


@pytest.mark.parametrize(
    "bad_value", (converters.UNSET_REQUIRED_VALUE, {}, [])
)
def test_bool_converter_raises(bad_value):
    with pytest.raises(Exception):
        converters.BoolConverter("foo").validate(bad_value)


@pytest.mark.parametrize(
    "value, expected", ((5, 5), ("5", 5), (None, None), (True, 1))
)
def test_int_converter(value, expected):
    assert expected == converters.IntConverter("foo").validate(value)


@pytest.mark.parametrize(
    "bad_value", ("3.14", converters.UNSET_REQUIRED_VALUE, "something", {}, [])
)
def test_int_converter_raises(bad_value):
    with pytest.raises(Exception):
        config.IntConverter("foo").validate(bad_value)



# Copyright 2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import collections
import string

import glom
import yaml

from klio_core import exceptions


class KlioConfigPreprocessor(object):
    """This handles everything between parsing the config from YAML and passing
    the parsed dict to KlioConfig."""

    # list of lambdas
    PLUGIN_PREPROCESSORS = []

    @classmethod
    def add_plugin_preprocessor(cls, proc):
        """proc should be a function that accepts the config_dict as a
        parameter and returns a modified version of it"""
        cls.PLUGIN_PREPROCESSORS.append(proc)

    @classmethod
    def _apply_plugin_preprocessors(cls, config_dict):
        for pre_processor in cls.PLUGIN_PREPROCESSORS:
            config_dict = pre_processor(config_dict)
        return config_dict

    @staticmethod
    def _transform_io_list(io_subsection_list):
        """Transform lists of dicts into a nested dict of dicts, where the keys
        for the top-level dict come from the `name` field in the nested dict.
        If `name` is not present, a name is auto-generated based on the index
        of the I/O and it's type.

        example:

        [
            {
                "name" : "my cool config",
                "type" : "bigquery",
                ....
            },
            {
                "type" : "file",
                ...
            }
        ]

        is transformed to

        {
            "my cool config" : {
                "name" : "my cool config",
                "type" : "bigquery",
                ...
            },
            "file0" : {
                "type" : "file",
                ...
            },
        }
        """

        type_counters = collections.defaultdict(int)
        io_dict = {}
        for conf in io_subsection_list:
            if "name" in conf:
                name = conf["name"]
                # TODO: right now "name" isn't supported in IOConfig (conflicts
                # with existing "name" attribute), once that's fixed we
                # shouldn't drop name here
                conf.pop("name")
            else:
                type_name = conf.get("type", "unknown")
                type_id = type_counters[type_name]
                type_counters[type_name] += 1
                name = "{}{}".format(type_name, type_id)

            io_dict[name] = conf

        return io_dict

    @classmethod
    def _transform_io_sections(cls, config_dict):
        clone = config_dict.copy()
        for io_type in ["events", "data"]:
            for io_direction in ["inputs", "outputs"]:
                path = "job_config.{}.{}".format(io_type, io_direction)

                glom_assign = glom.Assign(
                    path, glom.Spec((path, cls._transform_io_list))
                )
                glom.glom(
                    clone, glom_assign, default=None,
                )

        return clone

    @staticmethod
    def _apply_templates(raw_config_str, templates):
        """Applies templating to raw klio config.
        Example formats include:
        RAW = '{
            "allow_non_klio_messages": False,
            "events": {
                "inputs": {
                    "bigquery0": {
                        "type": "bigquery",
                        "table": "test.top_tracks_global",
                        "partition": "$YESTERDAY",
                    },
                    "bigquery1": {
                        "type": "bigquery",
                        "table": "test.top_artists_global",
                        "partition": "$TODAY",
                    },
                    "bigquery3": {
                        "type": "bigquery",
                        "table": "testing.${GENRE}_join_$COUNTRY",
                        "partition": "$TODAY",
                    }
                }
            }
        }'

        TEMPLATES = {
            "YESTERDAY": "12-31-2019",
            "TODAY": "01-01-2020",
            "GENRE": "ROCK",
            "COUNTRY": "STO"
        }

        Args:
            raw_config_str (str): raw klio config dict.
            Note mid string templating should be wrapped in ${variable}
            while lone or terminal variables can just be prefixed by a "$"
            templates (dict): template fields to apply to config dict
        Return:
            config_str (str): config dict as string with templates applied
        """
        template = string.Template(raw_config_str)
        try:
            return template.substitute(**templates)
        except KeyError as e:
            raise exceptions.KlioConfigTemplatingException(e)

    @staticmethod
    def _apply_overrides(raw_config, overrides):
        """Applies overrides to raw klio config.
        If a key already exists in the raw config,
        it will be updated with the override value provided in the overrides dict.
        If a key does not yet exist in the raw config,
        it will be created and assigned the override value.
        Example formats include:
            RAW = {
                "allow_non_klio_messages": False,
                "events": {
                    "inputs": {
                        "file0": {
                            "type": "file",
                            "location": "gs://sigint-output/yesterday.txt",
                        },
                        "file1": {
                            "type": "file",
                            "location": "gs://sigint-output/today.txt",
                        }
                    }
                }
            }

            OVER = {
                "allow_non_klio_messages": True,  # Non-nested key
                "events.inputs.file1.location": "gs://sigint-output/01-01-2020.txt",
                "events.inputs.file2.location": "gs://sigint-output/01-02-2020.txt",
                "events.inputs.file2.type": "file"
            }

        Args:
            raw_config (dict): raw klio config dict
            overrides (dict): override field to override value
        Return:
            config (dict): config dict with overrides applied
        """  # NOQA E501

        for path, value in overrides.items():
            glom.assign(raw_config, path, value, missing=dict)
        return raw_config

    @staticmethod
    def _parse_option_key_val_list(key_val_list):
        """Parses a list of strings in the form "<key>=<value>" into a dict of
        the form {"key" : "value"}"""
        key_vals = {}
        for key_val_str in key_val_list:
            # split "key=value" and trim whitespace from both
            pieces = ["".join(p.split()) for p in key_val_str.split("=")]
            if len(pieces) != 2:
                raise Exception(
                    (
                        "Malformed config override '{}',"
                        " should be '<key>=<value>'"
                    ).format(key_val_str)
                )
            key_vals[pieces[0].lower()] = pieces[1]
        return key_vals

    @classmethod
    def _parse_override_list(cls, raw_override_list):
        return cls._parse_option_key_val_list(raw_override_list)

    @classmethod
    def _parse_template_list(cls, raw_template_list):
        return cls._parse_option_key_val_list(raw_template_list)

    @classmethod
    def process(cls, raw_config_data, raw_template_list, raw_override_list):

        template_dict = cls._parse_template_list(raw_template_list)
        override_dict = cls._parse_override_list(raw_override_list)

        if isinstance(raw_config_data, dict):
            raw_config_data = yaml.dump(raw_config_data)

        templated_config_data = cls._apply_templates(
            raw_config_data, template_dict
        )

        config_dict = yaml.safe_load(templated_config_data)

        config_dict = cls._apply_plugin_preprocessors(config_dict)

        transformed_config = cls._transform_io_sections(config_dict)
        override_config = cls._apply_overrides(
            transformed_config, override_dict
        )

        return override_config



# Copyright 2019-2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import json
import os

import pytest

from klio_core import dataflow
from klio_core import utils


HERE = os.path.abspath(os.path.join(os.path.abspath(__file__), os.path.pardir))
FIXTURE_PATH = os.path.join(HERE, "fixtures")


@pytest.fixture
def mock_discovery_client(mocker, monkeypatch):
    mock = mocker.Mock()
    monkeypatch.setattr(dataflow.discovery, "build", lambda x, y: mock)
    return mock


def list_jobs():
    list_job_file = os.path.join(FIXTURE_PATH, "list_all_active_jobs.json")
    with open(list_job_file, "r") as f:
        return json.load(f)


def list_jobs_with_duplicates():
    list_job_file = os.path.join(
        FIXTURE_PATH, "list_all_active_jobs_duplicates.json"
    )
    with open(list_job_file, "r") as f:
        return json.load(f)


def get_job_detail():
    get_job_detail_file = os.path.join(FIXTURE_PATH, "get_job_detail.json")
    with open(get_job_detail_file, "r") as f:
        return json.load(f)


def get_job_detail_no_kind():
    return {"steps": [{"kind": "SomethingElse"}]}


def get_job_detail_no_username_value():
    return {
        "steps": [
            {
                "kind": "ParallelRead",
                "properties": {"username": "SomethingElse"},
            }
        ]
    }


def get_job_detail_no_topic():
    return {
        "steps": [
            {
                "kind": "ParallelRead",
                "properties": {"username": "ReadFromPubSub/Read"},
            }
        ]
    }


@pytest.mark.parametrize(
    "region,side_effect,exp_call_count,exp_log_count",
    (
        (None, [list_jobs_with_duplicates(), {}, {}], 13, 11),
        ("foo-region", [list_jobs(), {}, {}], 1, 0),
        (None, Exception("nojob4u"), 13, 13),
    ),
)
def test_find_job_by_name(
    region,
    side_effect,
    exp_call_count,
    exp_log_count,
    mock_discovery_client,
    caplog,
):
    client = dataflow.DataflowClient()
    locs = mock_discovery_client.projects.return_value.locations.return_value
    request = locs.jobs.return_value.list.return_value
    request.execute.side_effect = side_effect

    client.find_job_by_name("benchmark-beats", "gcp-project", region)

    assert exp_call_count == request.execute.call_count
    assert exp_log_count == len(caplog.records)


@pytest.mark.parametrize(
    "found_job,response,exp_log_count",
    (
        (True, Exception("foo"), 1),
        (True, get_job_detail(), 0),
        (False, None, 0),
    ),
)
def test_get_job_detail(
    found_job,
    response,
    exp_log_count,
    mock_discovery_client,
    caplog,
    mocker,
    monkeypatch,
):
    client = dataflow.DataflowClient()
    mock_find_job_by_name = mocker.Mock()
    monkeypatch.setattr(client, "find_job_by_name", mock_find_job_by_name)

    if found_job:
        mock_find_job_by_name.return_value = list_jobs().get("jobs")[0]
    else:
        mock_find_job_by_name.return_value = None

    locs = mock_discovery_client.projects.return_value.locations.return_value
    request = locs.jobs.return_value.get.return_value

    if isinstance(response, Exception):
        request.execute.side_effect = response
        exp_response = None
    else:
        request.execute.return_value = response
        exp_response = response

    ret_response = client.get_job_detail("job-name", "gcp-project")

    mock_find_job_by_name.assert_called_once_with(
        "job-name", "gcp-project", None
    )
    if found_job:
        assert 1 == request.execute.call_count
    assert exp_response == ret_response
    assert exp_log_count == len(caplog.records)


@pytest.mark.parametrize(
    "ret_job_info,exp_topic",
    (
        (lambda: None, None),
        (
            get_job_detail,
            "projects/sigint/topics/benchmark-audio-download-output",
        ),
        (get_job_detail_no_kind, None),
        (get_job_detail_no_username_value, None),
        (get_job_detail_no_topic, None),
    ),
)
def test_get_job_input_topic(
    ret_job_info, exp_topic, mock_discovery_client, mocker, monkeypatch
):
    client = dataflow.DataflowClient()
    mock_get_job_detail = mocker.Mock()
    monkeypatch.setattr(client, "get_job_detail", mock_get_job_detail)

    mock_get_job_detail.return_value = ret_job_info()

    ret_topic = client.get_job_input_topic("job-name", "gcp-project")

    assert exp_topic == ret_topic


@pytest.mark.parametrize(
    "api_version,exp_version", ((None, "v1b3"), ("v2", "v2"))
)
def test_get_dataflow_client(api_version, exp_version, mock_discovery_client):
    client = dataflow.get_dataflow_client(api_version)

    assert isinstance(client, dataflow.DataflowClient)
    global_client = utils.get_global("dataflow_client_%s" % exp_version)
    assert global_client == client

    delattr(utils, "klio_global_state_dataflow_client_%s" % exp_version)



# Copyright 2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from klio_core import config as kconfig
from klio_core import utils as core_utils


def assert_execution_success(result):
    """Helper for testing CLI commands that emits errors if execution failed"""
    if result.exception:
        if result.stdout:
            print("captured stdout: {}".format(result.stdout))
        raise result.exception

    assert 0 == result.exit_code


class MockKlioConfig(object):
    def __init__(self, module, mocker, monkeypatch, patch_os_getcwd):

        self.mock_get_config = mocker.patch.object(
            core_utils, "get_config_by_path"
        )
        self.mock_get_config_job_dir = mocker.Mock()
        monkeypatch.setattr(
            module.core_utils,
            "get_config_job_dir",
            self.mock_get_config_job_dir,
        )

        self.module = module
        self.patch_os_getcwd = patch_os_getcwd
        self.mocker = mocker
        self.monkeypatch = monkeypatch

    def setup(
        self,
        config_data,
        config_file,
        config_override=None,
        job_dir_override=None,
    ):
        self.config_override = config_override
        self.job_dir_override = job_dir_override
        self.config_data = config_data
        self.config_file = config_file

        self.effective_job_dir = self.job_dir_override or self.patch_os_getcwd
        self.effective_config_path = config_override or config_file

        self.mock_warn_if_py2_job = self.mocker.Mock()
        self.monkeypatch.setattr(
            self.module.core_utils,
            "warn_if_py2_job",
            self.mock_warn_if_py2_job,
        )

        self.mock_get_config_job_dir.return_value = (
            self.effective_job_dir,
            self.effective_config_path,
        )

        self.mock_get_config.return_value = config_data

        self.meta = core_utils.KlioConfigMeta(
            job_dir=self.effective_job_dir,
            config_file=config_override,
            config_path=self.effective_config_path,
        )

        self.klio_config = kconfig.KlioConfig(config_data)

        self.mock_klio_config = self.mocker.patch.object(
            core_utils.config, "KlioConfig"
        )
        self.mock_klio_config.return_value = self.klio_config

        return self.klio_config

    def assert_calls(self):
        self.mock_get_config_job_dir.assert_called_once_with(
            self.job_dir_override, self.config_override
        )
        self.mock_get_config.assert_called_once_with(
            self.effective_config_path
        )
        self.mock_klio_config.assert_called_once_with(
            self.config_data, raw_overrides=(), raw_templates=()
        )
        self.mock_warn_if_py2_job.assert_called_once_with(
            self.effective_job_dir
        )
